#!/usr/bin/env python
"""
A terminal browser, similar to less(1) for testing printable width of unicode.

This displays the full range of unicode points for 1 or 2-character wide
ideograms, with pipes ('|') that should always align for any terminal that
supports utf-8.

Interactive Keys:
  Navigation:
    k, y, UP          Scroll backward 1 line
    j, e, ENTER, DOWN Scroll forward 1 line
    f, SPACE, PGDOWN  Scroll forward 1 page
    b, PGUP           Scroll backward 1 page
    F, SHIFT-DOWN     Scroll forward 10 pages
    B, SHIFT-UP       Scroll backward 10 pages
    HOME              Go to top
    G, END            Go to bottom
    Ctrl-L            Refresh screen

  Mode Switching:
    0                 Exit VS mode (return to normal mode)
    1                 Narrow width (normal) / Narrow base filter (VS mode)
    2                 Wide width (normal) / Wide base filter (VS mode)
    5                 Switch to VS-15 mode (text style)
    6                 Switch to VS-16 mode (emoji style)
    c                 Toggle combining character mode
    w                 Toggle with/without variation selector (VS mode only)

  Display Adjustment:
    -, _              Decrease character name display length by 2
    +, =              Increase character name display length by 2
    v                 Select Unicode version

  Exit:
    q, Q              Quit browser

Note:
  Only one of --combining, --vs15, or --vs16 can be used at a time.
  The --without-vs option only applies when using --vs15 or --vs16.

  In VS mode, the display shows:
    - W/VS: Characters displayed with variation selector
    - WO/VS: Base characters displayed without variation selector
"""
# pylint: disable=C0103,W0622
#         Invalid constant name "echo"
#         Invalid constant name "flushout" (col 4)
#         Invalid module name "wcwidth-browser"

# std imports
import os
import sys
import signal
import string
import argparse
import functools
import unicodedata

# 3rd party
import blessed

# local
from wcwidth import ZERO_WIDTH, wcwidth, list_versions, _wcmatch_version

#: print function alias, does not end with line terminator.
echo = functools.partial(print, end='')
flushout = functools.partial(print, end='', flush=True)

#: printable length of highest unicode character description
LIMIT_UCS = 0x3fffd
UCS_PRINTLEN = len(f'{LIMIT_UCS:0x}')


def readline(term, width):
    """A rudimentary readline implementation."""
    text = ''
    while True:
        inp = term.inkey()
        if inp.code == term.KEY_ENTER:
            break
        if inp.code == term.KEY_ESCAPE:
            text = ''
            break
        if not inp.is_sequence and len(text) < width:
            text += inp
            echo(inp)
            flushout()
        elif inp.code in (term.KEY_BACKSPACE, term.KEY_DELETE):
            if text:
                text = text[:-1]
                echo('\b \b')
            flushout()
    return text


class WcWideCharacterGenerator:
    """Generator yields unicode characters of the given ``width``."""

    # pylint: disable=R0903
    #         Too few public methods (0/2)
    def __init__(self, width, unicode_version):
        """
        Class constructor.

        :param width: generate characters of given width.
        :param str unicode_version: Unicode Version for render.
        :type width: int
        """
        self.characters = (
            chr(idx) for idx in range(LIMIT_UCS)
            if wcwidth(chr(idx), unicode_version=unicode_version) == width)

    def __iter__(self):
        """Special method called by iter()."""
        return self

    def __next__(self):
        """Special method called by next()."""
        while True:
            ucs = next(self.characters)
            try:
                name = string.capwords(unicodedata.name(ucs))
            except ValueError:
                continue
            return (ucs, name)


class WcCombinedCharacterGenerator:
    """Generator yields unicode characters with combining."""

    # pylint: disable=R0903
    #         Too few public methods (0/2)

    def __init__(self, width, unicode_version):
        """
        Class constructor.

        :param int width: generate characters of given width.
        :param str unicode_version: Unicode version.
        """
        self.characters = []
        letters_o = ('o' * width)
        for (begin, end) in ZERO_WIDTH[_wcmatch_version(unicode_version)]:
            for val in [_val for _val in
                        range(begin, end + 1)
                        if _val <= LIMIT_UCS]:
                self.characters.append(
                    letters_o[:1] +
                    chr(val) +
                    letters_o[wcwidth(chr(val)) + 1:])
        self.characters.reverse()

    def __iter__(self):
        """Special method called by iter()."""
        return self

    def __next__(self):
        """
        Special method called by next().

        :return: unicode character and name, as tuple.
        :rtype: tuple[unicode, unicode]
        :raises StopIteration: no more characters
        """
        while True:
            if not self.characters:
                raise StopIteration
            ucs = self.characters.pop()
            try:
                name = string.capwords(unicodedata.name(ucs[1]))
            except ValueError:
                continue
            return (ucs, name)


class WcVariationSequenceGenerator:
    """Generator yields emoji variation sequences from emoji-variation-sequences.txt."""

    # pylint: disable=R0903
    #         Too few public methods (0/2)

    def __init__(self, base_width, unicode_version, variation_selector='VS15'):
        """
        Class constructor.

        :param int base_width: filter by base character width (1 or 2).
        :param str unicode_version: Unicode version.
        :param str variation_selector: 'VS15' or 'VS16'.
        """
        self.sequences = []

        # Determine which variation selector we're looking for
        vs_hex = 'FE0E' if variation_selector == 'VS15' else 'FE0F'

        # Find the emoji-variation-sequences.txt file
        script_dir = os.path.dirname(os.path.abspath(__file__))
        filepath = os.path.join(script_dir, '..', 'tests', 'emoji-variation-sequences.txt')

        with open(filepath, 'r', encoding='utf-8') as f:
            for line in f:
                # Skip comments and empty lines
                if line.startswith('#') or not line.strip():
                    continue

                # Only process lines with our target variation selector
                if vs_hex not in line:
                    continue

                # Parse line format: "0023 FE0E  ; text style;  # (1.1) NUMBER SIGN"
                parts = line.split(';')
                if len(parts) < 2:
                    continue

                codepoints = parts[0].strip().split()
                if len(codepoints) < 2:
                    continue

                try:
                    base_cp = int(codepoints[0], 16)
                    vs_cp = int(codepoints[1], 16)
                except ValueError:
                    continue

                # Check base character width matches our filter
                if wcwidth(chr(base_cp), unicode_version=unicode_version) != base_width:
                    continue

                # Extract name from comment
                comment_parts = line.split('#')
                if len(comment_parts) >= 2:
                    # Format: "# (1.1) NUMBER SIGN"
                    name_part = comment_parts[1].strip()
                    # Remove version info like "(1.1) "
                    if ')' in name_part:
                        name = name_part.split(')', 1)[1].strip()
                    else:
                        name = name_part
                    name = string.capwords(name)
                else:
                    name = "UNKNOWN"

                # Create the variation sequence
                sequence = chr(base_cp) + chr(vs_cp)
                self.sequences.append((sequence, name))

        self.sequences.reverse()

    def __iter__(self):
        """Special method called by iter()."""
        return self

    def __next__(self):
        """
        Special method called by next().

        :return: variation sequence and name, as tuple.
        :rtype: tuple[str, str]
        :raises StopIteration: no more sequences
        """
        if not self.sequences:
            raise StopIteration
        return self.sequences.pop()


class Style:
    """Styling decorator class instance for terminal output."""

    # pylint: disable=R0903
    #         Too few public methods (0/2)
    @staticmethod
    def attr_major(text):
        """Non-stylized callable for "major" text, for non-ttys."""
        return text

    @staticmethod
    def attr_minor(text):
        """Non-stylized callable for "minor" text, for non-ttys."""
        return text

    delimiter = '|'
    continuation = ' $'
    header_hint = '-'
    header_fill = '='
    name_len = 10
    alignment = 'right'

    def __init__(self, **kwargs):
        """
        Class constructor.

        Any given keyword arguments are assigned to the class attribute of the same name.
        """
        for key, val in kwargs.items():
            setattr(self, key, val)


class Screen:
    """Represents terminal style, data dimensions, and drawables."""

    intro_msg_fmt = ('Delimiters ({delim}) should align, '
                     'unicode version is {version}.')

    def __init__(self, term, style, wide=2):
        """Class constructor."""
        self.term = term
        self.style = style
        self.wide = wide

    @property
    def header(self):
        """Text of joined segments producing full heading."""
        return self.head_item * self.num_columns

    @property
    def hint_width(self):
        """Width of a column segment."""
        return sum((len(self.style.delimiter),
                    self.wide,
                    len(self.style.delimiter),
                    len(' '),
                    UCS_PRINTLEN + 2,
                    len(' '),
                    self.style.name_len,))

    @property
    def head_item(self):
        """Text of a single column heading."""
        delimiter = self.style.attr_minor(self.style.delimiter)
        hint = self.style.header_hint * self.wide
        heading = f'{delimiter}{hint}{delimiter}'

        def alignment(*args):
            if self.style.alignment == 'right':
                return self.term.rjust(*args)
            return self.term.ljust(*args)

        txt = alignment(heading, self.hint_width, self.style.header_fill)
        return self.style.attr_major(txt)

    def msg_intro(self, version):
        """Introductory message disabled above heading."""
        return self.term.center(self.intro_msg_fmt.format(
            delim=self.style.attr_minor(self.style.delimiter),
            version=self.style.attr_minor(version))).rstrip()

    @property
    def row_ends(self):
        """Bottom of page."""
        return self.term.height - 1

    @property
    def num_columns(self):
        """Number of columns displayed."""
        if self.term.is_a_tty:
            return self.term.width // self.hint_width
        return 1

    @property
    def num_rows(self):
        """Number of rows displayed."""
        return self.row_ends - self.row_begins - 1

    @property
    def row_begins(self):
        """Top row displayed for content."""
        # pylint: disable=R0201
        # Method could be a function (col 4)
        return 2

    @property
    def page_size(self):
        """Number of unicode text displayed per page."""
        return self.num_rows * self.num_columns


class Pager:
    """A less(1)-like browser for browsing unicode characters."""

    # pylint: disable=too-many-instance-attributes

    #: screen state for next draw method(s).
    STATE_CLEAN, STATE_DIRTY, STATE_REFRESH = 0, 1, 2

    def __init__(self, term, screen, character_factory, variation_selector=None,
                 show_variation_selector=True):
        """
        Class constructor.

        :param term: blessed Terminal class instance.
        :type term: blessed.Terminal
        :param screen: Screen class instance.
        :type screen: Screen
        :param character_factory: Character factory generator.
        :type character_factory: callable returning iterable.
        :param variation_selector: Variation selector mode ('VS15', 'VS16', or None).
        :type variation_selector: str or None
        :param show_variation_selector: Whether to display variation selector in VS mode.
        :type show_variation_selector: bool
        """
        self.term = term
        self.screen = screen
        self.character_factory = character_factory
        self.variation_selector = variation_selector
        self.show_variation_selector = show_variation_selector
        self.base_width_filter = screen.wide  # For VS mode filtering
        self.unicode_version = 'auto'
        self.dirty = self.STATE_REFRESH
        self.last_page = 0
        self._page_data = list()

    def on_resize(self, *args):
        """Signal handler callback for SIGWINCH."""
        # pylint: disable=W0613
        #         Unused argument 'args'
        assert self.term.width >= self.screen.hint_width, (
            f'Screen too small: {self.term.width}, must be at least {self.screen.hint_width}')
        self._set_lastpage()
        self.dirty = self.STATE_REFRESH

    def _set_lastpage(self):
        """Calculate value of class attribute ``last_page``."""
        self.last_page = (len(self._page_data) - 1) // self.screen.page_size

    def display_initialize(self):
        """Display 'please wait' message, and narrow build warning."""
        echo(self.term.home + self.term.clear)
        echo(self.term.move_y(self.term.height // 2))
        echo(self.term.center('Initializing page data ...').rstrip())
        flushout()

    def initialize_page_data(self):
        """Initialize the page data for the given screen."""
        # pylint: disable=attribute-defined-outside-init
        if self.term.is_a_tty:
            self.display_initialize()

        # Use variation sequence generator if in VS mode
        if self.variation_selector:
            self.character_generator = WcVariationSequenceGenerator(
                self.base_width_filter, self.unicode_version, self.variation_selector)
        else:
            self.character_generator = self.character_factory(
                self.screen.wide, self.unicode_version)

        self._page_data = list()
        while True:
            try:
                self._page_data.append(next(self.character_generator))
            except StopIteration:
                break
        self._set_lastpage()

    def page_data(self, idx, offset):
        """
        Return character data for page of given index and offset.

        :param idx: page index.
        :type idx: int
        :param offset: scrolling region offset of current page.
        :type offset: int
        :returns: list of tuples in form of ``(ucs, name)``
        :rtype: list[(unicode, unicode)]
        """
        size = self.screen.page_size

        while offset < 0 and idx:
            offset += size
            idx -= 1
        offset = max(0, offset)

        while offset >= size:
            offset -= size
            idx += 1

        if idx == self.last_page:
            offset = 0
        idx = min(max(0, idx), self.last_page)

        start = (idx * self.screen.page_size) + offset
        end = start + self.screen.page_size
        return (idx, offset), self._page_data[start:end]

    def _run_notty(self, writer):
        """Pager run method for terminals that are not a tty."""
        page_idx = page_offset = 0
        while True:
            npage_idx, _ = self.draw(writer, page_idx + 1, page_offset)
            if npage_idx == self.last_page:
                # page displayed was last page, quit.
                break
            page_idx = npage_idx
            self.dirty = self.STATE_DIRTY

    def _run_tty(self, writer, reader):
        """Pager run method for terminals that are a tty."""
        # allow window-change signal to reflow screen
        signal.signal(signal.SIGWINCH, self.on_resize)

        page_idx = page_offset = 0
        while True:
            if self.dirty:
                page_idx, page_offset = self.draw(writer,
                                                  page_idx,
                                                  page_offset)
                self.dirty = self.STATE_CLEAN
            inp = reader(timeout=0.25)
            if inp is not None:
                nxt, noff = self.process_keystroke(inp,
                                                   page_idx,
                                                   page_offset)
                if self.dirty:
                    continue
            if not self.dirty:
                self.dirty = nxt != page_idx or noff != page_offset
            page_idx, page_offset = nxt, noff
            if page_idx == -1:
                return

    def run(self, writer, reader):
        """
        Pager entry point.

        In interactive mode (terminal is a tty), run until ``process_keystroke()`` detects quit
        keystroke ('q').  In non-interactive mode, exit after displaying all unicode points.

        :param writer: callable writes to output stream, receiving unicode.
        :type writer: callable
        :param reader: callable reads keystrokes from input stream, sending instance of
            blessed.keyboard.Keystroke.
        :type reader: callable
        """
        self.initialize_page_data()
        if not self.term.is_a_tty:
            self._run_notty(writer)
        else:
            self._run_tty(writer, reader)

    def process_keystroke(self, inp, idx, offset):
        """
        Process keystroke ``inp``, adjusting screen parameters.

        :param inp: return value of blessed.Terminal.inkey().
        :type inp: blessed.keyboard.Keystroke
        :param idx: page index.
        :type idx: int
        :param offset: scrolling region offset of current page.
        :type offset: int
        :returns: tuple of next (idx, offset).
        :rtype: (int, int)
        """
        if inp.lower() in ('q', 'Q'):
            # exit
            return (-1, -1)
        self._process_keystroke_commands(inp)
        idx, offset = self._process_keystroke_movement(inp, idx, offset)
        return idx, offset

    def _process_keystroke_commands(self, inp):
        """Process keystrokes that issue commands (side effects)."""
        if inp in ('1', '2'):
            new_width = int(inp)
            if self.variation_selector:
                # In VS mode, change base width filter
                if self.base_width_filter != new_width:
                    self.base_width_filter = new_width
                    # If showing without VS, also update display width
                    if not self.show_variation_selector:
                        self.screen.wide = new_width
                    self.initialize_page_data()
                    self.on_resize(None, None)
            else:
                # In normal mode, change display width
                if self.screen.wide != new_width:
                    self.screen.wide = new_width
                    self.initialize_page_data()
                    self.on_resize(None, None)
        elif inp == '0':
            # Exit VS mode, return to normal mode
            if self.variation_selector:
                self.variation_selector = None
                # Keep current display width (screen.wide stays as is)
                self.initialize_page_data()
                self.on_resize(None, None)
        elif inp == '5':
            # Switch to VS-15 mode
            if self.variation_selector != 'VS15':
                self.variation_selector = 'VS15'
                self.base_width_filter = 1  # Default to narrow base
                # Display width depends on whether showing with or without VS
                if self.show_variation_selector:
                    self.screen.wide = 1  # VS-15 displays at width 1
                else:
                    self.screen.wide = self.base_width_filter  # Use base width
                self.initialize_page_data()
                self.on_resize(None, None)
        elif inp == '6':
            # Switch to VS-16 mode
            if self.variation_selector != 'VS16':
                self.variation_selector = 'VS16'
                self.base_width_filter = 1  # Default to narrow base
                # Display width depends on whether showing with or without VS
                if self.show_variation_selector:
                    self.screen.wide = 2  # VS-16 displays at width 2
                else:
                    self.screen.wide = self.base_width_filter  # Use base width
                self.initialize_page_data()
                self.on_resize(None, None)
        elif inp == 'c':
            # Switch on/off combining characters, clear VS mode
            self.variation_selector = None
            self.character_factory = (
                WcWideCharacterGenerator
                if self.character_factory != WcWideCharacterGenerator
                else WcCombinedCharacterGenerator)
            self.initialize_page_data()
            self.on_resize(None, None)
        elif inp == 'w':
            # Toggle showing variation selector (only in VS mode)
            if self.variation_selector:
                self.show_variation_selector = not self.show_variation_selector
                # Update display width based on whether we're showing VS or not
                if self.show_variation_selector:
                    # Showing with VS: use VS-determined width
                    self.screen.wide = 1 if self.variation_selector == 'VS15' else 2
                else:
                    # Showing without VS: use base character width
                    self.screen.wide = self.base_width_filter
                self.on_resize(None, None)
        elif inp in ('_', '-'):
            # adjust name length -2
            nlen = max(1, self.screen.style.name_len - 2)
            if nlen != self.screen.style.name_len:
                self.screen.style.name_len = nlen
                self.on_resize(None, None)
        elif inp in ('+', '='):
            # adjust name length +2
            nlen = min(self.term.width - 8, self.screen.style.name_len + 2)
            if nlen != self.screen.style.name_len:
                self.screen.style.name_len = nlen
                self.on_resize(None, None)
        elif inp == 'v':
            with self.term.location(x=0, y=self.term.height - 2):
                print(self.term.clear_eos())
                input_selection_msg = (
                    "--> Enter unicode version [{versions}] ("
                    "current: {self.unicode_version}):".format(
                        versions=', '.join(list_versions()),
                        self=self))
                echo('\n'.join(self.term.wrap(input_selection_msg,
                                              subsequent_indent='    ')))
                echo(' ')
                flushout()
                inp = readline(self.term, width=max(map(len, list_versions())))
                if inp.strip() and inp != self.unicode_version:
                    # set new unicode version -- page data must be
                    # re-initialized. Any version is legal, underlying
                    # library performs best-match (with warnings)
                    self.unicode_version = _wcmatch_version(inp)
                    self.initialize_page_data()
                self.on_resize(None, None)

    def _process_keystroke_movement(self, inp, idx, offset):
        """Process keystrokes that adjust index and offset."""
        term = self.term
        # a little vi-inspired.
        if inp in ('y', 'k') or inp.code in (term.KEY_UP,):
            # scroll backward 1 line
            offset -= self.screen.num_columns
        elif inp in ('e', 'j') or inp.code in (term.KEY_ENTER,
                                               term.KEY_DOWN,):
            # scroll forward 1 line
            offset = offset + self.screen.num_columns
        elif inp in ('f', ' ') or inp.code in (term.KEY_PGDOWN,):
            # scroll forward 1 page
            idx += 1
        elif inp == 'b' or inp.code in (term.KEY_PGUP,):
            # scroll backward 1 page
            idx = max(0, idx - 1)
        elif inp == 'F' or inp.code in (term.KEY_SDOWN,):
            # scroll forward 10 pages
            idx = max(0, idx + 10)
        elif inp == 'B' or inp.code in (term.KEY_SUP,):
            # scroll backward 10 pages
            idx = max(0, idx - 10)
        elif inp.code == term.KEY_HOME:
            # top
            idx, offset = (0, 0)
        elif inp == 'G' or inp.code == term.KEY_END:
            # bottom
            idx, offset = (self.last_page, 0)
        elif inp == '\x0c':
            self.dirty = True
        return idx, offset

    def draw(self, writer, idx, offset):
        """
        Draw the current page view to ``writer``.

        :param callable writer: callable writes to output stream, receiving unicode.
        :param int idx: current page index.
        :param int offset: scrolling region offset of current page.
        :returns: tuple of next (idx, offset).
        :rtype: (int, int)
        """
        # as our screen can be resized while we're mid-calculation,
        # our self.dirty flag can become re-toggled; because we are
        # not re-flowing our pagination, we must begin over again.
        while self.dirty:
            self.draw_heading(writer)
            self.dirty = self.STATE_CLEAN
            (idx, offset), data = self.page_data(idx, offset)
            for txt in self.page_view(data):
                writer(txt)
        self.draw_status(writer, idx)
        flushout()
        return idx, offset

    def draw_heading(self, writer):
        """
        Conditionally redraw screen when ``dirty`` attribute is valued REFRESH.

        When Pager attribute ``dirty`` is ``STATE_REFRESH``, cursor is moved to (0,0), screen is
        cleared, and heading is displayed.

        :param callable writer: callable writes to output stream, receiving unicode.
        :return: True if class attribute ``dirty`` is ``STATE_REFRESH``.
        :rtype: bool
        """
        if self.dirty == self.STATE_REFRESH:
            writer(''.join(
                (self.term.home, self.term.clear,
                 self.screen.msg_intro(version=self.unicode_version), '\n',
                 self.screen.header, '\n',)))
            return True
        return False

    def mode_label(self):
        """
        Return a label describing the current browsing mode.

        :return: Mode label string.
        :rtype: str
        """
        if self.variation_selector:
            # VS mode: show base width + VS type + with/without VS
            width_label = "NARROW" if self.base_width_filter == 1 else "WIDE"
            vs_display = "W/VS" if self.show_variation_selector else "WO/VS"
            return f"{width_label}+{self.variation_selector}+{vs_display}"
        elif self.character_factory == WcCombinedCharacterGenerator:
            # Combining mode
            return "COMBINING"
        else:
            # Normal mode: show display width
            return "NARROW" if self.screen.wide == 1 else "WIDE"

    def draw_status(self, writer, idx):
        """
        Conditionally draw status bar when output terminal is a tty.

        :param callable writer: callable writes to output stream, receiving unicode.
        :param int idx: current page position index.
        :type idx: int
        """
        if self.term.is_a_tty:
            writer(self.term.hide_cursor())
            style = self.screen.style
            writer(self.term.move(self.term.height - 1))
            if idx == self.last_page:
                last_end = '(END)'
            else:
                last_end = f'/{self.last_page}'

            # Get current mode label
            mode = self.mode_label()

            txt = ('Page {idx}{last_end} - [{mode}] - '
                   '{q} to quit, [keys: {keyset}]'
                   .format(idx=style.attr_minor(f'{idx}'),
                           last_end=style.attr_major(last_end),
                           mode=style.attr_major(mode),
                           keyset=style.attr_major('kjfbvc01256w-='),
                           q=style.attr_minor('q')))
            writer(self.term.center(txt).rstrip())

    def page_view(self, data):
        """
        Generator yields text to be displayed for the current unicode pageview.

        :param list[(unicode, unicode)] data: The current page's data as tuple     of ``(ucs,
        name)``.
        :returns: generator for full-page text for display
        """
        if self.term.is_a_tty:
            yield self.term.move(self.screen.row_begins, 0)
        # sequence clears to end-of-line
        clear_eol = self.term.clear_eol
        # sequence clears to end-of-screen
        clear_eos = self.term.clear_eos

        # track our current column and row, where column is
        # the whole segment of unicode value text, and draw
        # only self.screen.num_columns before end-of-line.
        #
        # use clear_eol at end of each row to erase over any
        # "ghosted" text, and clear_eos at end of screen to
        # clear the same, especially for the final page which
        # is often short.
        col = 0
        for ucs, name in data:
            val = self.text_entry(ucs, name)
            col += 1
            if col == self.screen.num_columns:
                col = 0
                if self.term.is_a_tty:
                    val = ''.join((val, clear_eol, '\n'))
                else:
                    val = ''.join((val.rstrip(), '\n'))
            yield val

        if self.term.is_a_tty:
            yield ''.join((clear_eol, '\n', clear_eos))

    def text_entry(self, ucs, name):
        """
        Display a single column segment row describing ``(ucs, name)``.

        :param str ucs: target unicode point character string.
        :param str name: name of unicode point.
        :return: formatted text for display.
        :rtype: unicode
        """
        style = self.screen.style
        if len(name) > style.name_len:
            idx = max(0, style.name_len - len(style.continuation))
            name = ''.join((name[:idx], style.continuation if idx else ''))
        if style.alignment == 'right':
            fmt = ' '.join(('0x{val:0>{ucs_printlen}x}',
                            '{name:<{name_len}s}',
                            '{delimiter}{ucs}{delimiter}'
                            ))
        else:
            fmt = ' '.join(('{delimiter}{ucs}{delimiter}',
                            '0x{val:0>{ucs_printlen}x}',
                            '{name:<{name_len}s}'))
        delimiter = style.attr_minor(style.delimiter)
        if len(ucs) != 1:
            # Variation sequence or combining character
            if self.variation_selector and not self.show_variation_selector:
                # VS mode, showing without variation selector - display only base
                val = ord(ucs[0])
                disp_ucs = style.attr_major(ucs[0])
            else:
                # Combining character or VS mode with variation selector shown
                val = ord(ucs[1])
                # a combining character displayed of any fg color
                # will reset the foreground character of the cell
                # combined with (iTerm2, OSX).
                disp_ucs = style.attr_major(ucs[0:2])
                if len(ucs) > 2:
                    disp_ucs += ucs[2]
        else:
            # non-combining
            val = ord(ucs)
            disp_ucs = style.attr_major(ucs)

        return fmt.format(name_len=style.name_len,
                          ucs_printlen=UCS_PRINTLEN,
                          delimiter=delimiter,
                          name=name,
                          ucs=disp_ucs,
                          val=val)


def validate_args(opts):
    """Validate result of parse_args() and return keyword arguments for main()."""
    if opts['--wide'] is None:
        opts['--wide'] = 2
    else:
        assert opts['--wide'] in ("1", "2"), opts['--wide']
    if opts['--alignment'] is None:
        opts['--alignment'] = 'left'
    else:
        assert opts['--alignment'] in ('left', 'right'), opts['--alignment']
    opts['--wide'] = int(opts['--wide'])

    # Ensure mutual exclusivity of --combining, --vs15, and --vs16
    exclusive_opts = [opts.get('--combining', False),
                      opts.get('--vs15', False),
                      opts.get('--vs16', False)]
    assert sum(bool(opt) for opt in exclusive_opts) <= 1, \
        "Only one of --combining, --vs15, or --vs16 can be used"

    # Set character factory and variation selector
    opts['character_factory'] = WcWideCharacterGenerator
    opts['variation_selector'] = None
    opts['base_width_filter'] = opts['--wide']  # Save base width filter
    opts['display_width'] = opts['--wide']  # Default display width
    opts['show_variation_selector'] = not opts.get('--without-vs', False)

    if opts.get('--combining'):
        opts['character_factory'] = WcCombinedCharacterGenerator
    elif opts.get('--vs15'):
        opts['variation_selector'] = 'VS15'
        # Display width depends on whether showing with or without VS
        if opts['show_variation_selector']:
            opts['display_width'] = 1  # VS-15 displays at width 1
        else:
            opts['display_width'] = opts['base_width_filter']  # Use base width
    elif opts.get('--vs16'):
        opts['variation_selector'] = 'VS16'
        # Display width depends on whether showing with or without VS
        if opts['show_variation_selector']:
            opts['display_width'] = 2  # VS-16 displays at width 2
        else:
            opts['display_width'] = opts['base_width_filter']  # Use base width

    return opts


def main(opts):
    """Program entry point."""
    term = blessed.Terminal()
    style = Style()

    # if the terminal supports colors, use a Style instance with some
    # standout colors (magenta, cyan).
    if term.number_of_colors:
        style = Style(attr_major=term.magenta,
                      attr_minor=term.bright_cyan,
                      alignment=opts['--alignment'])
    style.name_len = 10

    screen = Screen(term, style, wide=opts['display_width'])
    pager = Pager(term, screen, opts['character_factory'],
                  variation_selector=opts['variation_selector'],
                  show_variation_selector=opts['show_variation_selector'])

    # Set base width filter from command-line argument
    if opts['variation_selector']:
        pager.base_width_filter = opts['base_width_filter']

    with term.location(), term.cbreak(), \
            term.fullscreen(), term.hidden_cursor():
        pager.run(writer=echo, reader=term.inkey)
    return 0


def parse_args():
    """Parse command-line arguments using argparse."""
    # Extract description and usage from module docstring
    doc_lines = __doc__.split('\n')
    description = []
    for line in doc_lines:
        if line.strip() and not line.startswith('Usage:'):
            description.append(line)
        if line.startswith('Usage:'):
            break

    parser = argparse.ArgumentParser(
        description='A terminal browser for testing printable width of unicode.',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Interactive Keys:
  Navigation:
    k, y, UP          Scroll backward 1 line
    j, e, ENTER, DOWN Scroll forward 1 line
    f, SPACE, PGDOWN  Scroll forward 1 page
    b, PGUP           Scroll backward 1 page
    F, SHIFT-DOWN     Scroll forward 10 pages
    B, SHIFT-UP       Scroll backward 10 pages
    HOME              Go to top
    G, END            Go to bottom
    Ctrl-L            Refresh screen

  Mode Switching:
    0                 Exit VS mode (return to normal mode)
    1                 Narrow width (normal) / Narrow base filter (VS mode)
    2                 Wide width (normal) / Wide base filter (VS mode)
    5                 Switch to VS-15 mode (text style)
    6                 Switch to VS-16 mode (emoji style)
    c                 Toggle combining character mode
    w                 Toggle with/without variation selector (VS mode only)

  Display Adjustment:
    -, _              Decrease character name display length by 2
    +, =              Increase character name display length by 2
    v                 Select Unicode version

  Exit:
    q, Q              Quit browser

Notes:
  Only one of --combining, --vs15, or --vs16 can be used at a time.
  The --without-vs option only applies when using --vs15 or --vs16.

  In VS mode, the display shows:
    - W/VS: Characters displayed with variation selector
    - WO/VS: Base characters displayed without variation selector
""")

    parser.add_argument('--wide', metavar='<n>', type=str, default=None,
                        help='Browser 1 or 2 character-wide cells.')
    parser.add_argument('--alignment', metavar='<str>', type=str, default='left',
                        help='Choose left or right alignment. (default: left)')

    # Mutually exclusive group for mode selection
    mode_group = parser.add_mutually_exclusive_group()
    mode_group.add_argument('--combining', action='store_true',
                            help='Use combining character generator.')
    mode_group.add_argument('--vs15', action='store_true',
                            help='Browse emoji variation sequences with VS-15 (text style).')
    mode_group.add_argument('--vs16', action='store_true',
                            help='Browse emoji variation sequences with VS-16 (emoji style).')

    parser.add_argument('--without-vs', action='store_true',
                        help='Display base characters without variation selector.')

    args = parser.parse_args()

    # Convert to docopt-style dict format for compatibility with validate_args
    return {
        '--wide': args.wide,
        '--alignment': args.alignment,
        '--combining': args.combining,
        '--vs15': args.vs15,
        '--vs16': args.vs16,
        '--without-vs': args.without_vs,
        '--help': False,  # argparse handles this automatically
    }


if __name__ == '__main__':
    sys.exit(main(validate_args(parse_args())))



"""Tests for OSC 8 hyperlink parsing."""

# 3rd party
import pytest

# local
from wcwidth.hyperlink import Hyperlink, HyperlinkParams

PARAMS_PARSE_VALID = [
    ('\x1b]8;;http://example.com\x07', 'http://example.com', '', '\x07'),
    ('\x1b]8;id=a;http://example.com\x1b\\', 'http://example.com', 'id=a', '\x1b\\'),
]


@pytest.mark.parametrize('seq,url,params,term', PARAMS_PARSE_VALID)
def test_hyperlinkparams_parse_valid(seq, url, params, term):
    """Parse a valid OSC 8 open sequence."""
    result = HyperlinkParams.parse(seq)
    assert result is not None
    assert result.url == url
    assert result.params == params
    assert result.terminator == term


@pytest.mark.parametrize('seq', [
    'not an escape',
    '\x1b[31m',
    '',
])
def test_hyperlinkparams_parse_invalid(seq):
    """Parse an invalid/non-OSC-8 sequence returns None."""
    assert HyperlinkParams.parse(seq) is None


def test_hyperlinkparams_make_open():
    assert HyperlinkParams(url='http://example.com', params='id=a', terminator='\x07').make_open() == '\x1b]8;id=a;http://example.com\x07'


def test_hyperlinkparams_make_close():
    assert HyperlinkParams(url='http://example.com', terminator='\x07').make_close() == '\x1b]8;;\x07'


_HL = '\x1b]8;;http://example.com\x07Hello\x1b]8;;\x07'


def test_hyperlink_parse_valid():
    hl = Hyperlink.parse(_HL)
    assert hl is not None
    assert hl.text == 'Hello'
    assert hl.params.url == 'http://example.com'


@pytest.mark.parametrize('text,start', [
    ('Hello world', 0),
    ('\x1b[31mHello\x1b[0m', 0),   # SGR, not OSC 8
    ('\x1b]8;;http://example.com\x07Hello', 0),  # open without close
])
def test_hyperlink_parse_returns_none(text, start):
    assert Hyperlink.parse(text, start) is None


def test_hyperlink_find_close_not_found():
    assert Hyperlink.find_close('no escape here', 0) == (-1, -1)


def test_hyperlink_make_sequence():
    hl = Hyperlink.parse(_HL)
    assert hl is not None
    assert hl.make_sequence() == _HL


def test_hyperlink_display_width():
    hl = Hyperlink.parse(_HL)
    assert hl is not None
    assert hl.display_width() == 5



"""
Control character sets for terminal handling.

This module provides the control character sets used by the width() function to handle terminal
control characters.
"""

# Illegal C0/C1 control characters.
# These raise ValueError in 'strict' mode.
ILLEGAL_CTRL = frozenset(
    chr(c) for c in (
        list(range(0x01, 0x07)) +    # SOH, STX, ETX (^C), EOT (^D), ENQ, ACK
        list(range(0x10, 0x1b)) +    # DLE through SUB (^Z)
        list(range(0x1c, 0x20)) +    # FS, GS, RS, US
        [0x7f] +                      # DEL
        list(range(0x80, 0xa0))       # C1 control characters
    )
)

# Vertical movement control characters.
# These raise ValueError in 'strict' mode (indeterminate horizontal position).
VERTICAL_CTRL = frozenset({
    '\x0a',  # LF (line feed)
    '\x0b',  # VT (vertical tab)
    '\x0c',  # FF (form feed)
})

# Horizontal movement control characters.
# These affect cursor position and are tracked in 'strict' and 'parse' modes.
HORIZONTAL_CTRL = frozenset({
    '\x08',  # BS (backspace) - cursor left 1
    '\x09',  # HT (horizontal tab) - advance to next tab stop
    '\x0d',  # CR (carriage return) - cursor to column 0
})

# Terminal-valid zero-width control characters.
# These are allowed in all modes (zero-width, no movement).
ZERO_WIDTH_CTRL = frozenset({
    '\x00',  # NUL
    '\x07',  # BEL (bell)
    '\x0e',  # SO (shift out)
    '\x0f',  # SI (shift in)
})

# All control characters that need special handling (not regular printable).
ALL_CTRL = ILLEGAL_CTRL | VERTICAL_CTRL | HORIZONTAL_CTRL | ZERO_WIDTH_CTRL | {'\x1b'}



"""Tests for emoji width measurement and ZWJ sequences."""

# std imports
import os

# 3rd party
import pytest

# local
import wcwidth

# some tests cannot be done on some builds of python, where the internal
# unicode structure is limited to 0x10000 for memory conservation,
# "ValueError: unichr() arg not in range(0x10000) (narrow Python build)"
try:
    chr(0x2fffe)
    NARROW_ONLY = False
except ValueError:
    NARROW_ONLY = True


def make_sequence_from_line(line):
    # convert '002A FE0F  ; ..' -> (0x2a, 0xfe0f) -> chr(0x2a) + chr(0xfe0f)
    return ''.join(chr(int(cp, 16)) for cp in line.split(';', 1)[0].strip().split())


@pytest.mark.skipif(NARROW_ONLY, reason="Test cannot verify on python 'narrow' builds")
def emoji_zwj_sequence():
    """Emoji zwj sequence of four codepoints is just 2 cells."""
    phrase = ("\U0001f469"   # Base, Category So, East Asian Width property 'W' -- WOMAN
              "\U0001f3fb"   # Modifier, Category Sk, East Asian Width property 'W' -- EMOJI MODIFIER FITZPATRICK TYPE-1-2
              "\u200d"       # Joiner, Category Cf, East Asian Width property 'N'  -- ZERO WIDTH JOINER
              "\U0001f4bb")  # Fused, Category So, East Asian Width property 'W' -- PERSONAL COMPUTER
    # This test adapted from https://www.unicode.org/L2/L2023/23107-terminal-suppt.pdf
    expect_length_each = (2, 2, 0, 2)
    expect_length_phrase = 2

    # exercise,
    length_each = tuple(map(wcwidth.wcwidth, phrase))
    length_phrase = wcwidth.wcswidth(phrase)

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase


@pytest.mark.skipif(NARROW_ONLY, reason="Test cannot verify on python 'narrow' builds")
def test_unfinished_zwj_sequence():
    """Ensure index-out-of-bounds does not occur for ZWJ without any following character."""
    phrase = ("\U0001f469"   # Base, Category So, East Asian Width property 'W' -- WOMAN
              "\U0001f3fb"   # Modifier, Category Sk, East Asian Width property 'W' -- EMOJI MODIFIER FITZPATRICK TYPE-1-2
              "\u200d")      # Joiner, Category Cf, East Asian Width property 'N'  -- ZERO WIDTH JOINER
    expect_length_each = (2, 2, 0)
    expect_length_phrase = 2

    # exercise,
    length_each = tuple(map(wcwidth.wcwidth, phrase))
    length_phrase = wcwidth.wcswidth(phrase)

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase


@pytest.mark.skipif(NARROW_ONLY, reason="Test cannot verify on python 'narrow' builds")
def test_non_recommended_zwj_sequence():
    """Verify ZWJ with characters that cannot be joined, wcwidth does not verify."""
    phrase = ("\U0001f469"   # Base, Category So, East Asian Width property 'W' -- WOMAN
              "\U0001f3fb"   # Modifier, Category Sk, East Asian Width property 'W' -- EMOJI MODIFIER FITZPATRICK TYPE-1-2
              "\u200d")      # Joiner, Category Cf, East Asian Width property 'N'  -- ZERO WIDTH JOINER
    expect_length_each = (2, 2, 0)
    expect_length_phrase = 2

    # exercise,
    length_each = tuple(map(wcwidth.wcwidth, phrase))
    length_phrase = wcwidth.wcswidth(phrase)

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase


@pytest.mark.skipif(NARROW_ONLY, reason="Test cannot verify on python 'narrow' builds")
def test_another_emoji_zwj_sequence():
    phrase = (
        "\u26F9"        # PERSON WITH BALL
        "\U0001F3FB"    # EMOJI MODIFIER FITZPATRICK TYPE-1-2
        "\u200D"        # ZERO WIDTH JOINER
        "\u2640"        # FEMALE SIGN
        "\uFE0F")       # VARIATION SELECTOR-16
    expect_length_each = (1, 2, 0, 1, 0)
    expect_length_phrase = 2

    # exercise,
    length_each = tuple(map(wcwidth.wcwidth, phrase))
    length_phrase = wcwidth.wcswidth(phrase)

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase


@pytest.mark.skipif(NARROW_ONLY, reason="Test cannot verify on python 'narrow' builds")
def test_longer_emoji_zwj_sequence():
    """
    A much longer emoji ZWJ sequence of 10 total codepoints is just 2 cells!

    Also test the same sequence in duplicate, verifying multiple VS-16 sequences in a single
    function call.
    """
    # 'Category Code', 'East Asian Width property' -- 'description'
    phrase = ("\U0001F9D1"   # 'So', 'W' -- ADULT
              "\U0001F3FB"   # 'Sk', 'W' -- EMOJI MODIFIER FITZPATRICK TYPE-1-2
              "\u200d"       # 'Cf', 'N' -- ZERO WIDTH JOINER
              "\u2764"       # 'So', 'N' -- HEAVY BLACK HEART
              "\uFE0F"       # 'Mn', 'A' -- VARIATION SELECTOR-16
              "\u200d"       # 'Cf', 'N' -- ZERO WIDTH JOINER
              "\U0001F48B"   # 'So', 'W' -- KISS MARK
              "\u200d"       # 'Cf', 'N' -- ZERO WIDTH JOINER
              "\U0001F9D1"   # 'So', 'W' -- ADULT
              "\U0001F3FD"   # 'Sk', 'W' -- EMOJI MODIFIER FITZPATRICK TYPE-4
              ) * 2
    # This test adapted from https://www.unicode.org/L2/L2023/23107-terminal-suppt.pdf
    expect_length_each = (2, 2, 0, 1, 0, 0, 2, 0, 2, 2) * 2
    expect_length_phrase = 4

    # exercise,
    length_each = tuple(map(wcwidth.wcwidth, phrase))
    length_phrase = wcwidth.wcswidth(phrase)

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase


def read_sequences_from_file(filename):
    with open(os.path.join(os.path.dirname(__file__), filename), encoding='utf-8') as fp:
        lines = [line.strip()
                 for line in fp.readlines()
                 if not line.startswith('#') and line.strip()]
    sequences = [make_sequence_from_line(line) for line in lines]
    return lines, sequences


@pytest.mark.skipif(NARROW_ONLY, reason="Some sequences in text file are not compatible with 'narrow' builds")
@pytest.mark.skipif(not os.path.exists(os.path.join(os.path.dirname(__file__), 'emoji-zwj-sequences.txt')),
                    reason="emoji-zwj-sequences.txt; missing run bin/update-tables.py")
def test_recommended_emoji_zwj_sequences(benchmark):
    """Test wcswidth of all of the unicode.org-published emoji-zwj-sequences.txt."""
    lines, sequences = read_sequences_from_file('emoji-zwj-sequences.txt')

    def measure_all():
        errors = []
        for sequence, line in zip(sequences, lines):
            measured_width = wcwidth.wcswidth(sequence)
            if measured_width != 2:
                errors.append({
                    'expected_width': 2,
                    'line': line,
                    'measured_width': measured_width,
                    'sequence': sequence,
                })
        return errors

    errors = benchmark(measure_all)
    assert not errors
    assert len(sequences) >= 1468


@pytest.mark.skipif(not os.path.exists(os.path.join(os.path.dirname(__file__), 'emoji-variation-sequences.txt')),
                    reason="emoji-variation-sequences.txt is missing; run bin/update-tables.py")
def test_recommended_variation_16_sequences(benchmark):
    """Test wcswidth of all of the unicode.org-published emoji-variation-sequences.txt."""
    lines, sequences = read_sequences_from_file('emoji-variation-sequences.txt')
    vs16_sequences = [(seq, line) for seq, line in zip(sequences, lines) if '\ufe0f' in seq]

    def measure_all():
        errors = []
        for sequence, line in vs16_sequences:
            measured_width = wcwidth.wcswidth(sequence)
            if measured_width != 2:
                errors.append({
                    'expected_width': 2,
                    'line': line,
                    'measured_width': measured_width,
                    'sequence': sequence,
                })
        return errors

    errors = benchmark(measure_all)
    assert not errors
    assert len(sequences) >= 742


@pytest.mark.skipif(NARROW_ONLY, reason="Test cannot verify on python 'narrow' builds")
def test_regional_indicator_single():
    """Single Regional Indicator symbol is width 2."""
    assert wcwidth.wcwidth('\U0001F1FA') == 2
    assert wcwidth.wcswidth('\U0001F1FA') == 2


@pytest.mark.skipif(NARROW_ONLY, reason="Test cannot verify on python 'narrow' builds")
def test_regional_indicator_pair():
    """Flag pair (two Regional Indicators) is width 2, not 4."""
    assert wcwidth.wcswidth('\U0001F1FA\U0001F1F8') == 2


@pytest.mark.skipif(NARROW_ONLY, reason="Test cannot verify on python 'narrow' builds")
def test_regional_indicator_three():
    """Three Regional Indicators: one pair (2) + one single (2) = 4."""
    assert wcwidth.wcswidth('\U0001F1FA\U0001F1F8\U0001F1E6') == 4


@pytest.mark.skipif(NARROW_ONLY, reason="Test cannot verify on python 'narrow' builds")
def test_regional_indicator_four():
    """Four Regional Indicators: two pairs = 2 + 2 = 4."""
    assert wcwidth.wcswidth(
        '\U0001F1FA\U0001F1F8\U0001F1E6\U0001F1FA') == 4


@pytest.mark.skipif(NARROW_ONLY, reason="Test cannot verify on python 'narrow' builds")
def test_zwj_after_non_emoji():
    """ZWJ after non-emoji unconditionally consumes next character."""
    # This does *not* match most terminal behavior -- it is a negative test,
    # they fail because our library doesn't handle 'glitch' emoji as an
    # optimization. Non-emoji + ZWJ is undefined per Unicode UAX #29 GB11.
    assert wcwidth.wcswidth('xx\u200d\U0001F384') == 2
    assert wcwidth.wcswidth('a\u200d\U0001F600') == 1
    assert wcwidth.wcswidth('\u4e16\u200d\U0001F600') == 2


@pytest.mark.skipif(NARROW_ONLY, reason="Test cannot verify on python 'narrow' builds")
def test_fitzpatrick_standalone():
    """Standalone Fitzpatrick modifier is width 2."""
    assert wcwidth.wcwidth('\U0001F3FB') == 2
    assert wcwidth.wcswidth('\U0001F3FB') == 2


@pytest.mark.skipif(NARROW_ONLY, reason="Test cannot verify on python 'narrow' builds")
def test_fitzpatrick_after_emoji():
    """Fitzpatrick modifier after emoji base combines, total width 2."""
    assert wcwidth.wcswidth('\U0001F469\U0001F3FB') == 2


def test_vs16_effect():
    """Verify effect of VS-16 (always active with latest Unicode version)."""
    phrase = ("\u2640"        # FEMALE SIGN
              "\uFE0F")       # VARIATION SELECTOR-16

    expect_length_each = (1, 0)
    expect_length_phrase = 2

    # exercise,
    length_each = tuple(wcwidth.wcwidth(w_char) for w_char in phrase)
    length_phrase = wcwidth.wcswidth(phrase)

    # verify.
    assert length_each == expect_length_each
    assert length_phrase == expect_length_phrase



"""This is a python implementation of wcswidth()."""

from __future__ import annotations

from typing import Optional

__lazy_modules__ = [
    "wcwidth._constants",
    "wcwidth._wcwidth",
    "wcwidth.bisearch",
    "wcwidth.table_grapheme",
    "wcwidth.table_vs16",
]
# local
from . import table_grapheme_overrides
from ._wcwidth import wcwidth
from .bisearch import bisearch
from ._constants import (_EMOJI_ZWJ_SET,
                         _ISC_VIRAMA_SET,
                         _CATEGORY_MC_TABLE,
                         _FITZPATRICK_RANGE,
                         _REGIONAL_INDICATOR_SET,
                         resolve_terminal,
                         get_term_overrides)
from .table_vs15 import VS15_WIDE_TO_NARROW
from .table_vs16 import VS16_NARROW_TO_WIDE
from .table_grapheme import GRAPHEME_EXTEND


def _scan_zwj_cluster_end(text: str, start: int, end: int) -> int:
    """
    Scan forward from *start* (base character) to end of a ZWJ grapheme cluster.

    Follows the UAX #29 GB11 pattern (ExtPict Extend* ZWJ x ExtPict) chained repeatedly until no
    more ZWJ joins are found.
    """
    idx = start + 1
    # Skip Extend characters (Fitzpatrick modifiers, etc.) before first ZWJ
    while idx < end and bisearch(ord(text[idx]), GRAPHEME_EXTEND):
        idx += 1
    # Follow ZWJ chains
    while idx < end:
        if ord(text[idx]) != 0x200D:
            break
        idx += 1
        # GB11: \p{ExtPict} Extend* ZWJ × \p{ExtPict}
        # Extend modifiers (VS16, Fitzpatrick skin tones, etc.) attach to
        # the ExtPict *before* the ZWJ, not after it.  After ZWJ the next
        # codepoint is always an ExtPict directly, no Extend skip needed.
        if idx < end and ord(text[idx]) in _EMOJI_ZWJ_SET:
            idx += 1
            # Skip trailing Extend (VS16, etc.) after ExtPict before next ZWJ
            while idx < end and bisearch(ord(text[idx]), GRAPHEME_EXTEND):
                idx += 1
            continue
        break
    return idx


def wcswidth(
    pwcs: str,
    n: Optional[int] = None,
    unicode_version: str = 'auto',
    ambiguous_width: int = 1,
) -> int:
    """
    Given a unicode string, return its printable length on a terminal.

    See :ref:`Specification` for details of cell measurement.

    This implementation differs from Markus Khun's original POSIX C implementation, in that this
    ``wcswidth()`` processes graphemes strings yielded by :func:`wcwidth.iter_graphemes` defined by
    `Unicode Standard Annex #29`_. POSIX wcswidth(3) is not grapheme-aware and does not measure many
    kinds of Emojis or complex scripts correctly.

    :param pwcs: Measure width of given unicode string.
    :param n: When ``n`` is None (default), return the length of the entire
        string, otherwise only the first ``n`` characters are measured.
    :param unicode_version: Ignored. Retained for backwards compatibility.

        .. deprecated:: 0.3.0
           Only the latest Unicode version is now shipped.

    :param ambiguous_width: Width to use for East Asian Ambiguous (A)
        characters. Default is ``1`` (narrow). Set to ``2`` for CJK contexts.
    :returns: The width, in cells, needed to display the first ``n`` characters
        of the unicode string ``pwcs``.  Returns ``-1`` for C0 and C1 control
        characters!

    .. _`Unicode Standard Annex #29`: https://www.unicode.org/reports/tr29/
    """
    # pylint: disable=unused-argument,too-many-locals,too-many-statements,redefined-variable-type
    # pylint: disable=too-complex,too-many-branches,duplicate-code,too-many-nested-blocks

    # Fast path: pure ASCII printable strings are always width == length
    if n is None and pwcs.isascii() and pwcs.isprintable():
        return len(pwcs)

    _wcwidth = wcwidth if ambiguous_width == 1 else lambda c: wcwidth(c, 'auto', ambiguous_width)

    end = len(pwcs) if n is None else min(n, len(pwcs))
    total_width = 0
    idx = 0

    last_measured_idx = -2  # -2 sentinel blocks VS16/VS15 (no base available)
    last_measured_ucs = -1
    last_measured_w = 0
    prev_was_virama = False
    cluster_width = 0
    vs16_nw_table = VS16_NARROW_TO_WIDE['9.0.0']
    vs15_wn_table = VS15_WIDE_TO_NARROW['9.0.0']
    _bisearch = bisearch

    while idx < end:
        char = pwcs[idx]
        ucs = ord(char)

        # 5. ZWJ (U+200D): consumed without contributing width.
        # Virama codepoints are treated as zero-width combining marks (Mn). When a
        # virama+consonant sequence forms a conjunct, its width is capped at 2 cells.

        # ZWJ (U+200D)
        if ucs == 0x200D:
            if prev_was_virama:
                idx += 1
            elif idx + 1 < end:
                last_measured_w = 0
                prev_was_virama = False
                idx += 2
            else:
                prev_was_virama = False
                idx += 1
            continue

        # 6. VS16 (U+FE0F): converts preceding narrow character to wide.
        if ucs == 0xFE0F and last_measured_idx >= 0:
            if _bisearch(last_measured_ucs, vs16_nw_table):
                cluster_width = 2
            last_measured_idx = -2
            idx += 1
            continue

        # VS15 (U+FE0E): text variation selector, requests narrow presentation.
        if ucs == 0xFE0E and last_measured_idx >= 0:
            if bisearch(last_measured_ucs, vs15_wn_table) and last_measured_w == 2:
                total_width -= 1
            idx += 1
            continue

        # 7. Regional Indicator & Fitzpatrick (both above BMP)
        if ucs > 0xFFFF:
            if ucs in _REGIONAL_INDICATOR_SET:
                ri_before = 0
                j = idx - 1
                while j >= 0 and ord(pwcs[j]) in _REGIONAL_INDICATOR_SET:
                    ri_before += 1
                    j -= 1
                if ri_before % 2 == 1:
                    last_measured_ucs = ucs
                    idx += 1
                    continue
            elif (_FITZPATRICK_RANGE[0] <= ucs <= _FITZPATRICK_RANGE[1]
                  and last_measured_ucs in _EMOJI_ZWJ_SET):
                idx += 1
                continue

        # 8. Normal character: measure with wcwidth
        w = _wcwidth(char)
        if w < 0:
            return -1
        if w > 0:
            if prev_was_virama:
                cluster_width = 2
            elif cluster_width:
                total_width += cluster_width
                cluster_width = w
            else:
                cluster_width = w

            last_measured_idx = idx
            last_measured_ucs = ucs
            last_measured_w = w
            prev_was_virama = False
        elif ucs in _ISC_VIRAMA_SET:
            prev_was_virama = True
        elif last_measured_idx >= 0 and _bisearch(ucs, _CATEGORY_MC_TABLE):
            cluster_width = 2
            last_measured_idx = -2
            prev_was_virama = False
        else:
            prev_was_virama = False
        idx += 1

    if cluster_width:
        total_width += cluster_width
    return total_width


def wcstwidth(
    pwcs: str,
    n: Optional[int] = None,
    unicode_version: str = 'auto',
    ambiguous_width: int = 1,
    term_program: bool | str = True,
) -> int:
    """
    Given a unicode string, return its printable length on a terminal given by ``term_program``.

    See :ref:`Specification` for details of cell measurement.

    Unlike :func:`wcswidth`, this function applies per-terminal correction tables for
    emoji presentation and grapheme clusters.

    :param pwcs: Measure width of given unicode string.
    :param n: When ``n`` is None (default), return the length of the entire
        string, otherwise only the first ``n`` characters are measured.
    :param unicode_version: Ignored. Retained for backwards compatibility.
    :param ambiguous_width: Width to use for East Asian Ambiguous (A)
        characters. Default is ``1`` (narrow). Set to ``2`` for CJK contexts.
    :param term_program: Terminal software identifier for table correction.
        ``True`` (default) reads the ``TERM_PROGRAM`` or ``TERM`` environment
        variable for auto-detection.  ``False`` disables override lookup.
        Accepts a canonical terminal name matching :func:`list_term_programs`,
        such as from XTVERSION_, ENQ_, or ``TERM_PROGRAM``.

        .. versionadded:: 0.8.0
    :returns: The width, in cells, needed to display the first ``n`` characters
        of the unicode string ``pwcs``.  Returns ``-1`` for C0 and C1 control
        characters!
    """
    # pylint: disable=unused-argument,too-many-locals,too-many-statements,redefined-variable-type
    # pylint: disable=too-complex,too-many-branches,duplicate-code,too-many-nested-blocks
    # This function intentionally keeps all logic inline for performance.

    # Fast path: pure ASCII printable strings are always width == length
    if n is None and pwcs.isascii() and pwcs.isprintable():
        return len(pwcs)

    # Resolve terminal software for override lookup
    term_canonical = resolve_terminal(term_program)

    # Skip override lookup when no terminal detected (avoids lru_cache call overhead).
    # Extract locals for hot-loop performance (NamedTuple attribute access is slow).
    if term_canonical:
        overrides = get_term_overrides(term_canonical)
        _narrower = overrides.narrower
        _vs16_narrower = overrides.vs16_narrower
        _vs15_wider = overrides.vs15_wider
        _zeroer = overrides.zeroer
        _narrow_wider = overrides.narrow_wider
        _narrow_zeroer = overrides.narrow_zeroer
        _grapheme_overrides = table_grapheme_overrides.get(term_canonical)
    else:
        _narrower = ()
        _vs16_narrower = ()
        _vs15_wider = ()
        _zeroer = ()
        _narrow_wider = ()
        _narrow_zeroer = ()
        _grapheme_overrides = {}

    # Select wcwidth call pattern for best lru_cache performance
    _wcwidth = wcwidth if ambiguous_width == 1 else lambda c: wcwidth(c, 'auto', ambiguous_width)

    end = len(pwcs) if n is None else min(n, len(pwcs))
    total_width = 0
    idx = 0

    # grapheme-clustering state and local re-binding for performance.
    # Widths accumulate in cluster_width and flush at boundaries.  A cluster is a base character
    # plus combining marks, deferring the flush lets grapheme overrides replace the measured width
    # retrospectively.
    last_measured_idx = -2  # -2 sentinel blocks VS16/VS15 (no base available)
    last_measured_ucs = -1
    last_measured_w = 0
    prev_was_virama = False
    cluster_start = -1
    total_before_cluster = 0
    cluster_width = 0
    vs16_nw_table = VS16_NARROW_TO_WIDE['9.0.0']
    vs15_wn_table = VS15_WIDE_TO_NARROW['9.0.0']
    _bisearch = bisearch

    while idx < end:
        char = pwcs[idx]
        ucs = ord(char)

        #
        # Much of the logic below matches the logic in width(), but is repeated for improved
        # performance, they are given matching index reference numbers (starting at #5).
        #
        # 5. ZWJ (U+200D): consumed without contributing width.
        # Virama codepoints are treated as zero-width combining marks (Mn). When a
        # virama+consonant sequence forms a conjunct, its width is capped at 2 cells
        # matching behavior of popular terminals (PR #224)

        # ZWJ (U+200D)
        if ucs == 0x200D:
            if prev_was_virama:
                idx += 1
            elif idx + 1 < end:
                # Check for terminal grapheme override when base char is ExtPict/RI
                if (_grapheme_overrides
                        and last_measured_idx >= 0
                        and last_measured_ucs in _EMOJI_ZWJ_SET):
                    cluster_end = _scan_zwj_cluster_end(pwcs, last_measured_idx, end)
                    cluster = pwcs[last_measured_idx:cluster_end]
                    override_w = _grapheme_overrides.get(cluster)
                    if override_w is not None:
                        total_width += (override_w - last_measured_w)
                        last_measured_idx = -2
                        last_measured_ucs = -1
                        last_measured_w = 0
                        prev_was_virama = False
                        cluster_start = -1
                        idx = cluster_end
                        continue
                # No override; ZWJ breaks VS adjacency.
                # VS16 already set last_measured_idx = -2, blocking further VS16.
                last_measured_w = 0
                prev_was_virama = False
                idx += 2
            else:
                prev_was_virama = False
                idx += 1
            continue

        # 6. VS16 (U+FE0F): converts preceding narrow character to wide.
        if ucs == 0xFE0F and last_measured_idx >= 0:
            if _vs16_narrower and _bisearch(last_measured_ucs, _vs16_narrower):
                pass
            elif _bisearch(last_measured_ucs, vs16_nw_table):
                cluster_width = 2
            last_measured_idx = -2  # prevent double application
            idx += 1
            continue

        # VS15 (U+FE0E): text variation selector, requests narrow presentation.
        if ucs == 0xFE0E and last_measured_idx >= 0:
            base_ucs = last_measured_ucs
            vs15_narrow = bisearch(base_ucs, vs15_wn_table)
            if _vs15_wider and bisearch(base_ucs, _vs15_wider):
                vs15_narrow = False
            if vs15_narrow and last_measured_w == 2:
                total_width -= 1
            idx += 1
            continue

        # 7. Regional Indicator & Fitzpatrick (both above BMP)
        if ucs > 0xFFFF:
            if ucs in _REGIONAL_INDICATOR_SET:
                ri_before = 0
                j = idx - 1
                while j >= 0 and ord(pwcs[j]) in _REGIONAL_INDICATOR_SET:
                    ri_before += 1
                    j -= 1
                if ri_before % 2 == 1:
                    last_measured_ucs = ucs
                    idx += 1
                    continue
            elif (_FITZPATRICK_RANGE[0] <= ucs <= _FITZPATRICK_RANGE[1]
                  and last_measured_ucs in _EMOJI_ZWJ_SET):
                idx += 1
                continue

        # 8. Normal character: measure with wcwidth
        w = _wcwidth(char)
        if w < 0:
            # C0/C1 control character
            return -1
        # Apply single-codepoint terminal overrides (pre-merged tuples)
        if w == 2 and _narrower and bisearch(ucs, _narrower):
            w = 1
        elif w == 2 and _zeroer and bisearch(ucs, _zeroer):
            w = 0
        if w == 1 and _narrow_wider and bisearch(ucs, _narrow_wider):
            w = 2
        elif w == 1 and _narrow_zeroer and bisearch(ucs, _narrow_zeroer):
            w = 0
        if w > 0:
            # virama+consonant extends current cluster; otherwise start new
            if prev_was_virama:
                cluster_width = 2
            elif cluster_width:
                # flush previous cluster, check for grapheme overrides
                flushed = False
                if _grapheme_overrides and cluster_start >= 0:
                    # Two-phase override lookup: candidate (cluster+current) catches Lo+Lo pairs
                    # where both chars bear width (Thai KO KAI + SARA AM).  cluster_text (cluster
                    # alone) catches C+Mc clusters where the override key is shorter.
                    candidate = pwcs[cluster_start:idx + 1]
                    override_w = _grapheme_overrides.get(candidate)
                    if override_w is not None:
                        total_width = total_before_cluster + override_w
                        flushed = True
                        cluster_width = 0
                    else:
                        cluster_text = pwcs[cluster_start:idx]
                        override_w = _grapheme_overrides.get(cluster_text)
                        if override_w is not None:
                            total_width = total_before_cluster + override_w
                        else:
                            total_width += cluster_width
                else:
                    total_width += cluster_width
                if not flushed:
                    cluster_width = w
                    cluster_start = idx
                    total_before_cluster = total_width
            else:
                cluster_width = w
                cluster_start = idx
                total_before_cluster = total_width
            last_measured_idx = idx
            last_measured_ucs = ucs
            last_measured_w = w
            prev_was_virama = False
        elif ucs in _ISC_VIRAMA_SET:
            prev_was_virama = True
        elif last_measured_idx >= 0 and _bisearch(ucs, _CATEGORY_MC_TABLE):
            # Spacing Combining Mark (Mc) following a base character
            cluster_width = 2
            last_measured_idx = -2
            prev_was_virama = False
        else:
            prev_was_virama = False
        idx += 1

    if cluster_width:
        if _grapheme_overrides and cluster_start >= 0:
            cluster_text = pwcs[cluster_start:end]
            override_w = _grapheme_overrides.get(cluster_text)
            if override_w is not None:
                total_width = total_before_cluster + override_w
            else:
                total_width += cluster_width
        else:
            total_width += cluster_width
    return total_width



"""Performance benchmarks for wcwidth module."""

# std imports
import os
import sys
import unicodedata

# 3rd party
import pytest

# local
import wcwidth
# Access via import to trigger lazy-load on Python 3.15+; on older
# Python _width is already in sys.modules and this is a no-op.
import wcwidth._width as _width_module  # noqa: E402


def test_wcwidth_ascii(benchmark):
    """Benchmark wcwidth() with ASCII characters."""
    benchmark(wcwidth.wcwidth, 'A')


def test_wcwidth_wide(benchmark):
    """Benchmark wcwidth() with wide characters."""
    benchmark(wcwidth.wcwidth, '中')


def test_wcwidth_emoji(benchmark):
    """Benchmark wcwidth() with emoji."""
    benchmark(wcwidth.wcwidth, '😀')


def test_wcwidth_combining(benchmark):
    """Benchmark wcwidth() with combining characters."""
    benchmark(wcwidth.wcwidth, '\u0301')


def test_wcswidth_short_ascii(benchmark):
    """Benchmark wcswidth() with short ASCII string."""
    benchmark(wcwidth.wcswidth, 'hello')


def test_wcswidth_short_mixed(benchmark):
    """Benchmark wcswidth() with short mixed-width string."""
    benchmark(wcwidth.wcswidth, 'hello世界')


def test_wcswidth_long_ascii(benchmark):
    """Benchmark wcswidth() with long ASCII string."""
    text = 'The quick brown fox jumps over the lazy dog. ' * 10
    benchmark(wcwidth.wcswidth, text)


def test_wcswidth_long_japanese(benchmark):
    """Benchmark wcswidth() with long Japanese text."""
    text = 'コンニチハ、セカイ！' * 20
    benchmark(wcwidth.wcswidth, text)


def test_wcswidth_emoji_sequence(benchmark):
    """Benchmark wcswidth() with emoji sequences."""
    text = '👨\u200d👩\u200d👧\u200d👦' * 10
    benchmark(wcwidth.wcswidth, text)


# Regional Indicator benchmarks - paired flags and unpaired RI
RI_FLAGS_PAIRED = '🇺🇸🇬🇧🇫🇷🇩🇪🇯🇵' * 100
RI_FLAGS_UNPAIRED = '🇺🇸🇬🇧🇫' * 100


def test_wcswidth_ri_flags_paired(benchmark):
    """Benchmark wcswidth() with paired regional indicator flags."""
    benchmark(wcwidth.wcswidth, RI_FLAGS_PAIRED)


def test_wcswidth_ri_flags_unpaired(benchmark):
    """Benchmark wcswidth() with mixed paired and unpaired regional indicators."""
    benchmark(wcwidth.wcswidth, RI_FLAGS_UNPAIRED)


def test_width_ri_flags_paired(benchmark):
    """Benchmark width() with paired regional indicator flags."""
    benchmark(wcwidth.width, RI_FLAGS_PAIRED)


def test_width_ri_flags_unpaired(benchmark):
    """Benchmark width() with mixed paired and unpaired regional indicators."""
    benchmark(wcwidth.width, RI_FLAGS_UNPAIRED)


# NFC vs NFD comparison - text with combining marks
DIACRITICS_COMPOSED = 'café résumé naïve ' * 100
DIACRITICS_DECOMPOSED = unicodedata.normalize('NFD', DIACRITICS_COMPOSED)


def test_wcswidth_composed(benchmark):
    """Benchmark wcswidth() with NFC-composed text."""
    benchmark(wcwidth.wcswidth, DIACRITICS_COMPOSED)


def test_wcswidth_decomposed(benchmark):
    """Benchmark wcswidth() with NFD-decomposed text."""
    benchmark(wcwidth.wcswidth, DIACRITICS_DECOMPOSED)


def test_width_composed(benchmark):
    """Benchmark width() with NFC-composed text."""
    benchmark(wcwidth.width, DIACRITICS_COMPOSED)


def test_width_decomposed(benchmark):
    """Benchmark width() with NFD-decomposed text."""
    benchmark(wcwidth.width, DIACRITICS_DECOMPOSED)


def test_width_ascii(benchmark):
    """Benchmark width() with ASCII string."""
    benchmark(wcwidth.width, 'hello world')


def test_width_with_ansi_codes(benchmark):
    """Benchmark width() with ANSI escape sequences."""
    text = '\x1b[31mred text\x1b[0m'
    benchmark(wcwidth.width, text)


def test_width_complex_ansi(benchmark):
    """Benchmark width() with complex ANSI codes."""
    text = '\x1b[38;2;255;150;100mWARN\x1b[0m: Something happened'
    benchmark(wcwidth.width, text)


def test_iter_graphemes_ascii(benchmark):
    """Benchmark iter_graphemes() with ASCII string."""
    benchmark(lambda: list(wcwidth.iter_graphemes('hello world')))


def test_iter_graphemes_emoji(benchmark):
    """Benchmark iter_graphemes() with emoji."""
    text = 'ok🇿🇼café🏴󠁧󠁢󠁷󠁬󠁳󠁿'
    benchmark(lambda: list(wcwidth.iter_graphemes(text)))


def test_iter_graphemes_combining(benchmark):
    """Benchmark iter_graphemes() with combining characters."""
    text = 'café\u0301' * 10
    benchmark(lambda: list(wcwidth.iter_graphemes(text)))


def test_grapheme_boundary_before_short(benchmark):
    """Benchmark grapheme_boundary_before() near start of short string."""
    text = 'Hello 👋🏻!'
    benchmark(wcwidth.grapheme_boundary_before, text, 8)


def test_grapheme_boundary_before_long_end(benchmark):
    """Benchmark grapheme_boundary_before() near end of long line."""
    text = 'x' * 95 + '👨\u200d👩\u200d👧!'
    benchmark(wcwidth.grapheme_boundary_before, text, 100)


def test_grapheme_boundary_before_long_mid(benchmark):
    """Benchmark grapheme_boundary_before() in middle of long line."""
    text = 'x' * 50 + '👨\u200d👩\u200d👧' + 'y' * 50
    benchmark(wcwidth.grapheme_boundary_before, text, 55)


def test_iter_graphemes_reverse_short(benchmark):
    """Benchmark iter_graphemes_reverse() with short string."""
    text = 'café\u0301 🇫🇷!'
    benchmark(lambda: list(wcwidth.iter_graphemes_reverse(text)))


def test_iter_graphemes_reverse_long(benchmark):
    """Benchmark iter_graphemes_reverse() with long string."""
    text = 'The quick brown 🦊 jumps over the lazy 🐕. ' * 5
    benchmark(lambda: list(wcwidth.iter_graphemes_reverse(text)))


def test_ljust_ascii(benchmark):
    """Benchmark ljust() with ASCII string."""
    benchmark(wcwidth.ljust, 'hello', 20)


def test_ljust_japanese(benchmark):
    """Benchmark ljust() with Japanese characters."""
    benchmark(wcwidth.ljust, 'コンニチハ', 20)


def test_rjust_ascii(benchmark):
    """Benchmark rjust() with ASCII string."""
    benchmark(wcwidth.rjust, 'hello', 20)


def test_rjust_japanese(benchmark):
    """Benchmark rjust() with Japanese characters."""
    benchmark(wcwidth.rjust, 'コンニチハ', 20)


def test_center_ascii(benchmark):
    """Benchmark center() with ASCII string."""
    benchmark(wcwidth.center, 'hello', 20)


def test_center_mixed(benchmark):
    """Benchmark center() with mixed-width string."""
    benchmark(wcwidth.center, 'café\u0301', 20)


def test_wrap_short_ascii(benchmark):
    """Benchmark wrap() with short ASCII text."""
    text = 'The quick brown fox jumps over the lazy dog'
    benchmark(wcwidth.wrap, text, 20)


def test_wrap_long_text(benchmark):
    """Benchmark wrap() with long text."""
    text = 'The quick brown fox jumps over the lazy dog. ' * 20
    benchmark(wcwidth.wrap, text, 40)


def test_wrap_japanese(benchmark):
    """Benchmark wrap() with Japanese text."""
    text = 'コンニチハ、セカイ！これは日本語のテキストです。' * 5
    benchmark(wcwidth.wrap, text, 20)


def test_wrap_with_ansi(benchmark):
    """Benchmark wrap() with ANSI escape sequences."""
    text = '\x1b[31mThe quick brown fox\x1b[0m jumps over the lazy dog'
    benchmark(wcwidth.wrap, text, 20)


def test_wrap_with_ansi_no_propagate(benchmark):
    """Benchmark wrap() with ANSI but SGR propagation disabled."""
    text = '\x1b[31mThe quick brown fox\x1b[0m jumps over the lazy dog'
    benchmark(wcwidth.wrap, text, 20, propagate_sgr=False)


def test_wrap_complex_sgr(benchmark):
    """Benchmark wrap() with complex SGR (256-color, multiple attributes)."""
    text = '\x1b[1;3;38;5;208mBold italic orange text that wraps\x1b[0m'
    benchmark(wcwidth.wrap, text, 10)


def test_wrap_hyperlink_no_id(benchmark):
    """Benchmark wrap() with OSC 8 hyperlinks without id (requires id generation)."""
    link = '\x1b]8;;https://example.com/path\x1b\\click here for details\x1b]8;;\x1b\\'
    text = f'See {link} and also {link} for more. Read {link} now. ' * 10
    benchmark(wcwidth.wrap, text, 40)


def test_wrap_hyperlink_with_id(benchmark):
    """Benchmark wrap() with OSC 8 hyperlinks with existing ids."""
    link1 = '\x1b]8;id=a;https://example.com\x1b\\click here for details\x1b]8;;\x1b\\'
    link2 = '\x1b]8;id=b;https://other.org\x1b\\visit this page now\x1b]8;;\x1b\\'
    text = f'See {link1} and also {link2} for more. Read {link1} now. ' * 10
    benchmark(wcwidth.wrap, text, 40)


def test_wrap_hyperlink_mixed(benchmark):
    """Benchmark wrap() with mixed OSC 8 hyperlinks (with and without ids)."""
    no_id = '\x1b]8;;https://example.com\x1b\\link without id here\x1b]8;;\x1b\\'
    with_id = '\x1b]8;id=x;https://other.org\x1b\\link with id here\x1b]8;;\x1b\\'
    text = f'First {no_id} then {with_id} and {no_id} again {with_id} end. ' * 10
    benchmark(wcwidth.wrap, text, 40)


def test_clip_ascii(benchmark):
    """Benchmark clip() with ASCII string."""
    benchmark(wcwidth.clip, 'hello world', 0, 5)


def test_clip_japanese(benchmark):
    """Benchmark clip() with Japanese characters."""
    benchmark(wcwidth.clip, '中文字符串', 0, 5)


def test_clip_with_ansi(benchmark):
    """Benchmark clip() with ANSI sequences."""
    text = '\x1b[31m中文字\x1b[0m'
    benchmark(wcwidth.clip, text, 0, 3)


def test_clip_with_ansi_no_propagate(benchmark):
    """Benchmark clip() with ANSI but SGR propagation disabled."""
    text = '\x1b[31m中文字\x1b[0m'
    benchmark(wcwidth.clip, text, 0, 3, propagate_sgr=False)


def test_clip_complex_sgr(benchmark):
    """Benchmark clip() with complex SGR clipping from middle."""
    text = '\x1b[1;38;5;208mHello world text\x1b[0m'
    benchmark(wcwidth.clip, text, 6, 11)


def test_clip_long_cjk_past_window(benchmark):
    """Benchmark clip() with long CJK text, narrow window (early-exit path)."""
    text = '中文测试字符串' * 100  # 700 chars, no escape sequences
    benchmark(wcwidth.clip, text, 0, 50)


def test_clip_dense_ansi_past_window(benchmark):
    """Benchmark clip() with dense ANSI sequences past clip window (SGR tracking)."""
    text = '\x1b[31mred\x1b[0m \x1b[32mgreen\x1b[0m \x1b[33myellow\x1b[0m ' * 50
    benchmark(wcwidth.clip, text, 6, 30)


def test_clip_dense_ansi_no_propagate(benchmark):
    """Benchmark clip() with dense ANSI sequences, SGR propagation disabled."""
    text = '\x1b[31mred\x1b[0m \x1b[32mgreen\x1b[0m \x1b[33myellow\x1b[0m ' * 50
    benchmark(wcwidth.clip, text, 6, 30, propagate_sgr=False)


def test_clip_osc8_hyperlinks(benchmark):
    """Benchmark clip() with OSC 8 hyperlinks (hyperlink parsing path)."""
    text = '\x1b]8;;http://example.com\x07Click Here\x1b]8;;\x07 ' * 20
    benchmark(wcwidth.clip, text, 0, 80)


def test_width_osc66(benchmark):
    """Benchmark width() with OSC 66 text sizing sequences."""
    text = '\x1b]66;w=2;XY\x07\x1b]66;s=3;ABC\x07'
    benchmark(wcwidth.width, text)


def test_clip_osc66(benchmark):
    """Benchmark clip() with OSC 66 text sizing sequences."""
    text = '\x1b]66;w=2;XY\x07\x1b]66;s=3;ABC\x07'
    benchmark(wcwidth.clip, text, 3, 8)


def test_clip_cursor_cr_overwrite(benchmark):
    """Benchmark clip() with carriage-return overwrite (painter path)."""
    text = 'hello\rworld ' * 20
    benchmark(wcwidth.clip, text, 0, 50)


def test_clip_cursor_csi_backward(benchmark):
    """Benchmark clip() with CSI cursor-backward sequences (painter path)."""
    text = 'hello\x1b[2Dxy ' * 20
    benchmark(wcwidth.clip, text, 0, 40)


def test_clip_long_ascii_fastpath(benchmark):
    """Benchmark clip() with long ASCII string (fast-path slice)."""
    text = 'hello world ' * 1000
    benchmark(wcwidth.clip, text, 500, 600)


def test_clip_with_ansi_no_overtype(benchmark):
    """Benchmark clip() with ANSI sequences, overtyping disabled."""
    text = '\x1b[31m中文字\x1b[0m'
    benchmark(wcwidth.clip, text, 0, 3, overtyping=False)


def test_clip_complex_sgr_no_overtype(benchmark):
    """Benchmark clip() with complex SGR, overtyping disabled."""
    text = '\x1b[1;38;5;208mHello world text\x1b[0m'
    benchmark(wcwidth.clip, text, 6, 11, overtyping=False)


def test_clip_dense_ansi_no_overtype(benchmark):
    """Benchmark clip() with dense ANSI, overtyping disabled."""
    text = '\x1b[31mred\x1b[0m \x1b[32mgreen\x1b[0m \x1b[33myellow\x1b[0m ' * 50
    benchmark(wcwidth.clip, text, 6, 30, overtyping=False)


def test_clip_dense_ansi_no_propagate_no_overtype(benchmark):
    """Benchmark clip() with dense ANSI, SGR propagation and overtyping disabled."""
    text = '\x1b[31mred\x1b[0m \x1b[32mgreen\x1b[0m \x1b[33myellow\x1b[0m ' * 50
    benchmark(wcwidth.clip, text, 6, 30, propagate_sgr=False, overtyping=False)


def test_clip_dense_ansi_overtype(benchmark):
    """Benchmark clip() with dense ANSI, overtyping forced (painter path)."""
    text = '\x1b[31mred\x1b[0m \x1b[32mgreen\x1b[0m \x1b[33myellow\x1b[0m ' * 50
    benchmark(wcwidth.clip, text, 6, 30, overtyping=True)


def test_clip_long_cjk_overtype(benchmark):
    """Benchmark clip() with long CJK, overtyping forced (painter path)."""
    text = '中文测试字符串' * 100
    benchmark(wcwidth.clip, text, 0, 50, overtyping=True)


def test_width_dense_ansi_control_codes_ignore(benchmark):
    """Benchmark width() with dense ANSI and control_codes='ignore'."""
    text = '\x1b[31mred\x1b[0m \x1b[32mgreen\x1b[0m \x1b[33myellow\x1b[0m ' * 50
    benchmark(wcwidth.width, text, control_codes='ignore')


def test_width_complex_ansi_control_codes_ignore(benchmark):
    """Benchmark width() with complex ANSI and control_codes='ignore'."""
    text = '\x1b[38;2;255;150;100mWARN\x1b[0m: \x1b[1mBold\x1b[0m \x1b[4mUnderline\x1b[0m'
    benchmark(wcwidth.width, text, control_codes='ignore')


def test_clip_dense_ansi_control_codes_ignore(benchmark):
    """Benchmark clip() with dense ANSI, control_codes='ignore' (skips painter/OSC)."""
    text = '\x1b[31mred\x1b[0m \x1b[32mgreen\x1b[0m \x1b[33myellow\x1b[0m ' * 50
    benchmark(wcwidth.clip, text, 6, 30, control_codes='ignore')


def test_clip_long_cjk_control_codes_ignore(benchmark):
    """Benchmark clip() with long CJK and control_codes='ignore' (early-exit path)."""
    text = '中文测试字符串' * 100
    benchmark(wcwidth.clip, text, 0, 50, control_codes='ignore')


def test_clip_cursor_cr_control_codes_ignore(benchmark):
    """Benchmark clip() with CR overwrite and control_codes='ignore' (painter skipped)."""
    text = 'hello\rworld ' * 20
    benchmark(wcwidth.clip, text, 0, 50, control_codes='ignore')


def test_clip_dense_ansi_no_propagate_control_codes_ignore(benchmark):
    """Benchmark clip() with dense ANSI, propagate_sgr=False and control_codes='ignore'."""
    text = '\x1b[31mred\x1b[0m \x1b[32mgreen\x1b[0m \x1b[33myellow\x1b[0m ' * 50
    benchmark(wcwidth.clip, text, 6, 30, propagate_sgr=False, control_codes='ignore')


def test_clip_long_ascii_control_codes_ignore(benchmark):
    """Benchmark clip() with long ASCII and control_codes='ignore' (fast-path slice)."""
    text = 'hello world ' * 1000
    benchmark(wcwidth.clip, text, 500, 600, control_codes='ignore')


def test_wrap_with_ansi_control_codes_ignore(benchmark):
    """Benchmark wrap() with ANSI sequences and control_codes='ignore'."""
    text = '\x1b[31mThe quick brown fox jumps over the lazy dog.\x1b[0m Did it really? ' * 20
    benchmark(wcwidth.wrap, text, 40, control_codes='ignore')


def test_ljust_ascii_control_codes_ignore(benchmark):
    """Benchmark ljust() with ASCII and control_codes='ignore'."""
    benchmark(wcwidth.ljust, 'hello', 20, control_codes='ignore')


def test_propagate_sgr_multiline(benchmark):
    """Benchmark propagate_sgr() with multiple lines."""
    lines = ['\x1b[1;31mline one', 'line two', 'line three\x1b[0m']
    benchmark(wcwidth.propagate_sgr, lines)


def test_propagate_sgr_no_sequences(benchmark):
    """Benchmark propagate_sgr() fast path (no sequences)."""
    lines = ['line one', 'line two', 'line three']
    benchmark(wcwidth.propagate_sgr, lines)


def test_strip_sequences_simple(benchmark):
    """Benchmark strip_sequences() with simple ANSI codes."""
    text = '\x1b[31mred\x1b[0m'
    benchmark(wcwidth.strip_sequences, text)


def test_strip_sequences_complex(benchmark):
    """Benchmark strip_sequences() with complex ANSI codes."""
    text = '\x1b[38;2;255;150;100mWARN\x1b[0m: \x1b[1mBold\x1b[0m \x1b[4mUnderline\x1b[0m'
    benchmark(wcwidth.strip_sequences, text)


def test_iter_sequences_plain(benchmark):
    """Benchmark iter_sequences() with plain text."""
    benchmark(lambda: list(wcwidth.iter_sequences('hello world')))


def test_iter_sequences_mixed(benchmark):
    """Benchmark iter_sequences() with mixed content."""
    text = '\x1b[31mred\x1b[0m normal \x1b[32mgreen\x1b[0m'
    benchmark(lambda: list(wcwidth.iter_sequences(text)))


# Brahmic script benchmarks -- text with virama conjuncts
BRAHMIC_DEVANAGARI = 'हिन्दी भाषा में लिखा गया पाठ है। क्षत्रिय स्त्री ' * 20
BRAHMIC_BENGALI = 'বাংলা ভাষায় লেখা একটি পাঠ। বাঙ্গালী ভাষা ' * 20


def test_wcswidth_brahmic_devanagari(benchmark):
    """Benchmark wcswidth() with Devanagari text containing conjuncts."""
    benchmark(wcwidth.wcswidth, BRAHMIC_DEVANAGARI)


def test_wcswidth_brahmic_bengali(benchmark):
    """Benchmark wcswidth() with Bengali text containing conjuncts."""
    benchmark(wcwidth.wcswidth, BRAHMIC_BENGALI)


def test_width_brahmic_devanagari(benchmark):
    """Benchmark width() with Devanagari text containing conjuncts."""
    benchmark(wcwidth.width, BRAHMIC_DEVANAGARI)


def test_width_brahmic_bengali(benchmark):
    """Benchmark width() with Bengali text containing conjuncts."""
    benchmark(wcwidth.width, BRAHMIC_BENGALI)


# UDHR-based benchmarks,
# Load combined text (500+ world languages)
UDHR_FILE = os.path.join(os.path.dirname(__file__), 'udhr_combined.txt')
UDHR_TEXT = ''
UDHR_LINES = []
UDHR_WIDTHS = []
UDHR_FILLCHAR = '█'
UDHR_SAMPLE_EVERY = 20
if os.path.exists(UDHR_FILE):
    with open(UDHR_FILE, encoding='utf-8') as f:
        UDHR_TEXT = f.read()
    _all_lines = [line.rstrip() for line in UDHR_TEXT.splitlines() if line.strip()]
    UDHR_LINES = [l for i, l in enumerate(_all_lines) if i % UDHR_SAMPLE_EVERY == 0]
    UDHR_TEXT = '\n'.join(UDHR_LINES)
    UDHR_WIDTHS = [wcwidth.width(line) for line in UDHR_LINES]

_udhr_skip = pytest.mark.skipif(
    not os.path.exists(UDHR_FILE),
    reason=f"{os.path.basename(UDHR_FILE)} is missing; run bin/update-tables.py",
)

_py38_skip_pedantic = pytest.mark.skipif(
    sys.version_info[:2] < (3, 9),
    reason='benchmark.pedantic() not supported in python 3.8 or earlier')


@_udhr_skip
@_py38_skip_pedantic
def test_wrap_udhr(benchmark):
    """Benchmark wrap() with multilingual UDHR text."""
    if not hasattr(benchmark, 'pedantic'):
        pytest.skip('pytest-codspeed not installed')
    result = benchmark.pedantic(wcwidth.wrap, args=(UDHR_TEXT, 80), rounds=1, iterations=1)
    assert len(result)
    assert all(0 <= wcwidth.width(_l) <= 80 for _l in result)


@_udhr_skip
@_py38_skip_pedantic
def test_width_udhr(benchmark):
    """Benchmark width() with multilingual UDHR text."""
    if not hasattr(benchmark, 'pedantic'):
        pytest.skip('pytest-codspeed not installed')
    result = benchmark.pedantic(wcwidth.width, args=(UDHR_TEXT,), rounds=1, iterations=1)
    assert result > 0


@_udhr_skip
@_py38_skip_pedantic
def test_width_udhr_lines(benchmark):
    """Benchmark width() on individual UDHR lines."""
    if not hasattr(benchmark, 'pedantic'):
        pytest.skip('pytest-codspeed not installed')
    result = benchmark.pedantic(lambda: sum(wcwidth.width(line) for line in UDHR_LINES),
                                rounds=1, iterations=1)
    assert result > 0


@_udhr_skip
@_py38_skip_pedantic
def test_width_wcswidth_consistency_udhr(benchmark):
    """Verify width() and wcswidth() agree for printable multilingual text."""

    def check():
        failures = []
        for line in UDHR_LINES:
            if not line or not line.isprintable():
                continue
            w = wcwidth.width(line)
            wcs = wcwidth.wcswidth(line)
            if w != wcs:
                failures.append((line[:60], w, wcs))
        return failures
    if not hasattr(benchmark, 'pedantic'):
        pytest.skip('pytest-codspeed not installed')
    failures = benchmark.pedantic(check, rounds=1, iterations=1)
    assert not failures


@_udhr_skip
@_py38_skip_pedantic
def test_width_fastpath_integrity_udhr(benchmark):
    """Verify width() produces identical results with and without the fast path."""
    saved = _width_module._WIDTH_FAST_PATH_MIN_LEN

    def check():
        _width_module._WIDTH_FAST_PATH_MIN_LEN = 0
        fast_total = sum(wcwidth.width(line) for line in UDHR_LINES)
        _width_module._WIDTH_FAST_PATH_MIN_LEN = 999_999
        parse_total = sum(wcwidth.width(line) for line in UDHR_LINES)
        return fast_total, parse_total

    if not hasattr(benchmark, 'pedantic'):
        pytest.skip('pytest-codspeed not installed')
    fast_total, parse_total = benchmark.pedantic(check, rounds=1, iterations=1)
    _width_module._WIDTH_FAST_PATH_MIN_LEN = saved
    assert fast_total == parse_total


@_udhr_skip
@_py38_skip_pedantic
def test_ljust_udhr_lines(benchmark):
    """Benchmark ljust() on UDHR lines."""
    if not hasattr(benchmark, 'pedantic'):
        pytest.skip('pytest-codspeed not installed')
    benchmark.pedantic(lambda: [wcwidth.ljust(line, w + 1, UDHR_FILLCHAR)
                                for line, w in zip(UDHR_LINES, UDHR_WIDTHS)],
                       rounds=1, iterations=1)


_TERM_PROGRAMS = [
    'ghostty',
    'xterm.js',
]


@pytest.mark.parametrize('term_program', _TERM_PROGRAMS)
def test_wcstwidth_term_program(benchmark, term_program):
    """Benchmark wcstwidth() with term_program (ghostty=0 overrides vs xterm.js=237)."""
    text = 'Hello 世界 😀 café ' * 20
    benchmark(wcwidth.wcstwidth, text, term_program=term_program)


@pytest.mark.parametrize('term_program', _TERM_PROGRAMS)
def test_wcstwidth_ri_term_program(benchmark, term_program):
    """Benchmark wcstwidth() with RI flags and term_program."""
    benchmark(wcwidth.wcstwidth, RI_FLAGS_PAIRED, term_program=term_program)


@pytest.mark.parametrize('term_program', _TERM_PROGRAMS)
def test_wcstwidth_emoji_term_program(benchmark, term_program):
    """Benchmark wcstwidth() with emoji ZWJ sequences and term_program."""
    text = '👨\u200d👩\u200d👧\u200d👦' * 20
    benchmark(wcwidth.wcstwidth, text, term_program=term_program)


@pytest.mark.parametrize('term_program', _TERM_PROGRAMS)
def test_wcstwidth_wide_term_program(benchmark, term_program):
    """Benchmark wcstwidth() with wide CJK and term_program."""
    text = 'コンニチハ、セカイ！' * 20
    benchmark(wcwidth.wcstwidth, text, term_program=term_program)


@pytest.mark.parametrize('term_program', _TERM_PROGRAMS)
def test_width_term_program(benchmark, term_program):
    """Benchmark width() with term_program (ghostty=0 overrides vs xterm.js=237)."""
    text = 'Hello 世界 😀 café ' * 20
    benchmark(wcwidth.width, text, term_program=term_program)


@pytest.mark.parametrize('term_program', _TERM_PROGRAMS)
def test_width_ri_term_program(benchmark, term_program):
    """Benchmark width() with RI flags and term_program."""
    benchmark(wcwidth.width, RI_FLAGS_PAIRED, term_program=term_program)


# VS16/VS15-heavy text to exercise the vs16_narrower/vs15_wider bisearch paths
_VS16_TEXT = ('\u263A\uFE0F'      # WHITE SMILING FACE + VS16
              '\u2764\uFE0F'       # HEAVY BLACK HEART + VS16
              '\u2600\uFE0F'       # BLACK SUN WITH RAYS + VS16
              '\u2615\uFE0F') * 25  # HOT BEVERAGE + VS16

_VS15_TEXT = ('\u2615\uFE0E'       # HOT BEVERAGE + VS15 (wide=2 narrows to 1)
              '\u231A\uFE0E'       # WATCH + VS15 (wide=2 narrows to 1)
              '\u23F0\uFE0E') * 34  # ALARM CLOCK + VS15 (wide=2 narrows to 1)


@pytest.mark.parametrize('term_program', _TERM_PROGRAMS)
def test_wcstwidth_vs16_term_program(benchmark, term_program):
    """Benchmark wcswidth() with VS16 sequences to exercise vs16_narrower bisearch."""
    benchmark(wcwidth.wcstwidth, _VS16_TEXT, term_program=term_program)


@pytest.mark.parametrize('term_program', _TERM_PROGRAMS)
def test_wcstwidth_vs15_term_program(benchmark, term_program):
    """Benchmark wcswidth() with VS15 sequences to exercise vs15_wider bisearch."""
    benchmark(wcwidth.wcstwidth, _VS15_TEXT, term_program=term_program)


@pytest.mark.parametrize('term_program', _TERM_PROGRAMS)
def test_width_vs16_term_program(benchmark, term_program):
    """Benchmark width() with VS16 sequences to exercise vs16_narrower bisearch."""
    benchmark(wcwidth.width, _VS16_TEXT, term_program=term_program)


@pytest.mark.parametrize('term_program', _TERM_PROGRAMS)
def test_width_vs15_term_program(benchmark, term_program):
    """Benchmark width() with VS15 sequences to exercise vs15_wider bisearch."""
    benchmark(wcwidth.width, _VS15_TEXT, term_program=term_program)



"""Tests for grapheme cluster segmentation."""

# std imports
import os

# 3rd party
import pytest

# local
from wcwidth import iter_graphemes, iter_graphemes_reverse, grapheme_boundary_before

try:
    chr(0x2fffe)
    NARROW_ONLY = False
except ValueError:
    NARROW_ONLY = True


def parse_grapheme_break_test_line(line):
    """Parse a line from GraphemeBreakTest.txt."""
    data, _, _ = line.partition('#')
    data = data.strip()
    if not data:
        return None, None

    parts = []
    current_cluster = []

    for token in data.split():
        if token == '÷':
            if current_cluster:
                parts.append(current_cluster)
                current_cluster = []
        elif token == '×':
            pass
        else:
            try:
                current_cluster.append(int(token, 16))
            except ValueError:
                continue

    if current_cluster:
        parts.append(current_cluster)

    all_codepoints = []
    expected_clusters = []
    for cluster in parts:
        cluster_str = ''.join(chr(cp) for cp in cluster)
        expected_clusters.append(cluster_str)
        all_codepoints.extend(cluster)

    if not all_codepoints:
        return None, None

    input_str = ''.join(chr(cp) for cp in all_codepoints)
    return input_str, expected_clusters


def read_grapheme_break_test():
    """Read and parse GraphemeBreakTest.txt."""
    test_file = os.path.join(os.path.dirname(__file__), 'GraphemeBreakTest.txt')
    if not os.path.exists(test_file):
        return []

    test_cases = []
    with open(test_file, encoding='utf-8') as f:
        for line_num, line in enumerate(f, 1):
            line = line.strip()
            if not line or line.startswith('#'):
                continue
            input_str, expected = parse_grapheme_break_test_line(line)
            if input_str is not None:
                test_cases.append(pytest.param(input_str, expected, id=f"line{line_num}"))

    return test_cases


@pytest.mark.parametrize(("input_str", "expected"), [
    ('', []),
    ('a', ['a']),
    ('abc', ['a', 'b', 'c']),
    ('cafe\u0301', ['c', 'a', 'f', 'e\u0301']),
    ('\r\n', ['\r\n']),
    ('ok\r\nok', ['o', 'k', '\r\n', 'o', 'k']),
    ('\r', ['\r']),
    ('ok\rok', ['o', 'k', '\r', 'o', 'k']),
    ('\n', ['\n']),
    ('ok\nok', ['o', 'k', '\n', 'o', 'k']),
    ('\r\r', ['\r', '\r']),
    ('ok\r\rok', ['o', 'k', '\r', '\r', 'o', 'k']),
])
def test_core_grapheme(input_str, expected):
    """Basic grapheme cluster segmentation."""
    assert list(iter_graphemes(input_str)) == expected


@pytest.mark.parametrize(("input_str", "start", "end", "expected"), [
    ('abcdef', 2, None, ['c', 'd', 'e', 'f']),
    ('abcdef', 0, 4, ['a', 'b', 'c', 'd']),
    ('abcdef', 1, 4, ['b', 'c', 'd']),
    ('abc', 10, None, []),
    ('abc', 0, 10, ['a', 'b', 'c']),
])
def test_iter_graphemes_slice(input_str, start, end, expected):
    """Grapheme iteration with start/end parameters."""
    assert list(iter_graphemes(input_str, start=start, end=end)) == expected


HANGUL_LV = '\u1100\u1161'
HANGUL_LVT = '\uAC00\u11A8'
FLAG_US = '\U0001F1FA\U0001F1F8'
FLAG_AU = '\U0001F1E6\U0001F1FA'
RI_A = '\U0001F1E6'
FAMILY = '\U0001F468\u200D\U0001F469\u200D\U0001F467'
WAVE_SKIN = '\U0001F44B\U0001F3FB'
HEART_EMOJI = '\u2764\uFE0F'


@pytest.mark.skipif(NARROW_ONLY, reason="requires wide Unicode")
@pytest.mark.parametrize(("input_str", "expected"), [
    (HANGUL_LV, [HANGUL_LV]),
    ('ok' + HANGUL_LV + 'ok', ['o', 'k', HANGUL_LV, 'o', 'k']),
    (HANGUL_LVT, [HANGUL_LVT]),
    ('ok' + HANGUL_LVT + 'ok', ['o', 'k', HANGUL_LVT, 'o', 'k']),
    (FLAG_US, [FLAG_US]),
    ('ok' + FLAG_US + 'ok', ['o', 'k', FLAG_US, 'o', 'k']),
    (FLAG_US + RI_A, [FLAG_US, RI_A]),
    ('ok' + FLAG_US + RI_A + 'ok', ['o', 'k', FLAG_US, RI_A, 'o', 'k']),
    (FLAG_US + FLAG_AU, [FLAG_US, FLAG_AU]),
    ('ok' + FLAG_US + FLAG_AU + 'ok', ['o', 'k', FLAG_US, FLAG_AU, 'o', 'k']),
    (FAMILY, [FAMILY]),
    ('ok' + FAMILY + 'ok', ['o', 'k', FAMILY, 'o', 'k']),
    (WAVE_SKIN, [WAVE_SKIN]),
    ('ok' + WAVE_SKIN + 'ok', ['o', 'k', WAVE_SKIN, 'o', 'k']),
    (HEART_EMOJI, [HEART_EMOJI]),
    ('ok' + HEART_EMOJI + 'ok', ['o', 'k', HEART_EMOJI, 'o', 'k']),
])
def test_wide_unicode_graphemes(input_str, expected):
    """Grapheme segmentation for wide Unicode characters."""
    assert list(iter_graphemes(input_str)) == expected


@pytest.mark.skipif(NARROW_ONLY, reason="requires wide Unicode")
@pytest.mark.skipif(not os.path.exists(os.path.join(os.path.dirname(__file__), 'GraphemeBreakTest.txt')),
                    reason="GraphemeBreakTest.txt is missing; run bin/update-tables.py")
@pytest.mark.parametrize(("input_str", "expected"), read_grapheme_break_test())
def test_unicode_grapheme_break_test(input_str, expected):
    """Validate against official Unicode GraphemeBreakTest.txt."""
    assert list(iter_graphemes(input_str)) == expected


# Prepend: Arabic Number Sign
PREPEND_CHAR = '\u0600'
# Multiple combining marks: e + acute + grave
MULTI_COMBINE = 'e\u0301\u0300'


# grapheme_boundary_before(text, pos) returns start of grapheme cluster before pos.
# (text, pos, expected): pos=search from here, expected=where cluster starts
@pytest.mark.parametrize(("text", "pos", "expected"), [
    # 'abc': 0=a, 1=b, 2=c
    ('abc', 3, 2),  # from end -> 'c' at 2
    ('abc', 2, 1),  # from 'c' -> 'b' at 1
    ('abc', 1, 0),  # from 'b' -> 'a' at 0
    # 'a\r\nb': CRLF is one cluster (GB3)
    ('a\r\nb', 3, 1),  # from 'b' -> '\r\n' at 1
    # 'café': e + combining acute is one cluster (GB9)
    ('cafe\u0301', 5, 3),  # from end -> 'é' at 3
    ('cafe\u0301', 4, 3),  # from acute -> still 'é' at 3
    # Multiple combining marks: e + acute + grave (GB9)
    ('a' + MULTI_COMBINE + 'b', 4, 1),  # from 'b' -> e+marks at 1
    # Prepend + char is one cluster (GB9b)
    (PREPEND_CHAR + 'a', 2, 0),  # whole cluster
    # Prepend + Control: control breaks (GB4)
    (PREPEND_CHAR + '\n', 2, 1),  # '\n' separate at 1
    # C1 control (NEL, 0x85) stops backward scan in _find_cluster_start (GB4)
    ('X\x85\u0301', 3, 2),
])
def test_grapheme_boundary_before_basic(text, pos, expected):
    """Basic grapheme_boundary_before tests."""
    assert grapheme_boundary_before(text, pos) == expected


@pytest.mark.skipif(NARROW_ONLY, reason="requires wide Unicode")
@pytest.mark.parametrize(("text", "pos", "expected"), [
    # 'Hi 👋🏻!': 0=H,1=i,2=space,3=wave,4=skin,5=!; wave+skin is one cluster
    ('Hi \U0001F44B\U0001F3FB!', 6, 5),  # from end -> '!' at 5
    ('Hi \U0001F44B\U0001F3FB!', 5, 3),  # from '!' -> wave+skin at 3
    ('Hi \U0001F44B\U0001F3FB!', 3, 2),  # from wave -> space at 2
    # 'a🇺🇸b': 0=a,1-2=flag,3=b; flag is one cluster (GB12/13)
    ('a' + FLAG_US + 'b', 4, 3),  # from end -> 'b' at 3
    ('a' + FLAG_US + 'b', 3, 1),  # from 'b' -> flag at 1
    # Three RIs (🇺🇸🇦): flag + solo RI
    (FLAG_US + RI_A, 3, 2),  # from end -> solo RI at 2
    (FLAG_US + RI_A, 2, 0),  # from solo -> flag at 0
    # 'a👨‍👩‍👧b': 0=a,1-5=family,6=b; ZWJ sequence is one cluster (GB11)
    ('a' + FAMILY + 'b', 7, 6),  # from end -> 'b' at 6
    ('a' + FAMILY + 'b', 6, 1),  # from 'b' -> family at 1
])
def test_grapheme_boundary_before_unicode(text, pos, expected):
    """grapheme_boundary_before with emoji and wide Unicode."""
    assert grapheme_boundary_before(text, pos) == expected


@pytest.mark.parametrize(("input_str", "expected"), [
    ('', []),
    ('abc', ['c', 'b', 'a']),
    # café with combining mark mixed with CRLF
    ('cafe\u0301\r\nok', ['k', 'o', '\r\n', 'e\u0301', 'f', 'a', 'c']),
])
def test_iter_graphemes_reverse_basic(input_str, expected):
    """Basic iter_graphemes_reverse tests."""
    assert list(iter_graphemes_reverse(input_str)) == expected


@pytest.mark.skipif(NARROW_ONLY, reason="requires wide Unicode")
@pytest.mark.parametrize(("input_str", "expected"), [
    # Multiple emoji types in one string
    ('cafe\u0301 ' + WAVE_SKIN + ' ' + FLAG_US + '!',
     ['!', FLAG_US, ' ', WAVE_SKIN, ' ', 'e\u0301', 'f', 'a', 'c']),
    # Two families
    (FAMILY + FAMILY, [FAMILY, FAMILY]),
    # Flag + solo RI + text
    ('Hi' + FLAG_US + RI_A + '!', ['!', RI_A, FLAG_US, 'i', 'H']),
])
def test_iter_graphemes_reverse_unicode(input_str, expected):
    """iter_graphemes_reverse with wide Unicode."""
    assert list(iter_graphemes_reverse(input_str)) == expected


@pytest.mark.skipif(NARROW_ONLY, reason="requires wide Unicode")
@pytest.mark.skipif(not os.path.exists(os.path.join(os.path.dirname(__file__), 'GraphemeBreakTest.txt')),
                    reason="GraphemeBreakTest.txt is missing; run bin/update-tables.py")
@pytest.mark.parametrize(("input_str", "expected"), read_grapheme_break_test())
def test_grapheme_roundtrip_consistency(input_str, expected):
    """Forward and reverse iteration produce identical boundaries."""
    forward = list(iter_graphemes(input_str))
    reverse = list(iter_graphemes_reverse(input_str))[::-1]
    assert forward == reverse


def test_grapheme_boundary_before_edge_cases():
    """Edge cases for grapheme_boundary_before."""
    assert grapheme_boundary_before('abc', 0) == 0
    assert grapheme_boundary_before('abc', 100) == 2  # pos > len clamps
    assert grapheme_boundary_before('', 0) == 0


def test_iter_graphemes_reverse_edge_cases():
    """Edge cases for iter_graphemes_reverse."""
    assert list(iter_graphemes_reverse('abcdef', start=2, end=5)) == ['e', 'd', 'c']
    assert list(iter_graphemes_reverse('abc', start=0, end=100)) == ['c', 'b', 'a']
    assert not list(iter_graphemes_reverse('abc', start=5))
    assert not list(iter_graphemes_reverse('abc', start=2, end=2))
    # PREPEND + char is one grapheme (GB9b), so start=1 yields nothing (won't split)
    assert not list(iter_graphemes_reverse(PREPEND_CHAR + 'a', start=1))
    # But start=0 yields the full grapheme
    assert list(iter_graphemes_reverse(PREPEND_CHAR + 'a', start=0)) == [PREPEND_CHAR + 'a']
    # Negative start is clamped to 0
    assert list(iter_graphemes_reverse('abc', start=-5)) == ['c', 'b', 'a']



"""Tests for Text Sizing Protocol (OSC 66) support."""

# 3rd party
import pytest

# local
from wcwidth import (TextSizing,
                     TextSizingParams,
                     clip,
                     width,
                     wcswidth,
                     iter_sequences,
                     strip_sequences)
from wcwidth.text_sizing import TEXT_FIELD_MAPPING
from wcwidth.escape_sequences import TEXT_SIZING_PATTERN

_W_HI = TEXT_FIELD_MAPPING['w'].high
_N_HI = TEXT_FIELD_MAPPING['n'].high
_D_HI = TEXT_FIELD_MAPPING['d'].high

CONTROL_CODES_PARAMS_CASES = [
    ('x=2', "", "Unknown text sizing field 'x' in "),
    ('s=3:x=3', "s=3", "Unknown text sizing field 'x' in "),
    ('s=2:x=3:w=9', f"s=2:w={_W_HI}", "Unknown text sizing field 'x' in "),
    ('xyz=2', "", "Unknown text sizing field 'xyz' in "),
    ('xxx', "", "Expected '=' in text sizing parameter"),
    ('s=xxx', "", "Illegal text sizing value 'xxx' in "),
    ('s=-99', "", "Out of bounds text sizing value '-99' in "),
    ('s=99', f"s={_W_HI}", "Out of bounds text sizing value '99' in "),
    ('w=-1', "", "Out of bounds text sizing value '-1' in "),
    ('w=8', f"w={_W_HI}", "Out of bounds text sizing value '8' in "),
    ('n=20', f"n={_N_HI}", "Out of bounds text sizing value '20' in "),
    ('d=99', f"d={_D_HI}", "Out of bounds text sizing value '99' in "),
    ('v=5', "v=2", "Out of bounds text sizing value '5' in "),
    ('h=3', "h=2", "Out of bounds text sizing value '3' in "),
]


@pytest.mark.parametrize('given_params,expected_remainder,expected_exc,', CONTROL_CODES_PARAMS_CASES)
def test_text_sizing_params_control_codes(given_params, expected_remainder, expected_exc):
    """Verify control_codes='strict' and 'parse' behavior in TextSizingParams.from_params()."""
    # assert control_codes='strict' raises expected exception,
    with pytest.raises(ValueError) as exc_info:
        TextSizingParams.from_params(given_params, control_codes='strict')
    assert exc_info.value.args[0].startswith(expected_exc)

    # when 'parse' (default), any illegal argument or value is filtered, excluded, or clipped
    params = TextSizingParams.from_params(given_params)
    assert params.make_sequence() == expected_remainder


@pytest.mark.parametrize('given_params,expected_remainder,expected_exc,', CONTROL_CODES_PARAMS_CASES)
def test_text_sizing_width_control_codes(given_params, expected_remainder, expected_exc):
    """Verify control_codes='strict' with invalid OSC 66 sequences in wciwdth.width()."""
    seq1 = '\x1b]66;' + given_params + ';ABC' + '\x07'
    seq2 = '\x1b]66;' + given_params + ';ABC' + '\x1b\\'
    for seq in (seq1, seq2):
        with pytest.raises(ValueError) as exc_info:
            width(seq, control_codes='strict')
        assert exc_info.value.args[0].startswith(expected_exc)


@pytest.mark.parametrize('params,expected_repr', [
    (TextSizingParams(), 'TextSizingParams()'),
    (TextSizingParams(scale=2, width=1), 'TextSizingParams(scale=2, width=1)'),
    (TextSizingParams(scale=2, width=3, numerator=1, denominator=2,
                      vertical_align=1, horizontal_align=2),
     'TextSizingParams(scale=2, width=3, numerator=1, denominator=2, '
     'vertical_align=1, horizontal_align=2)'),
])
def test_text_sizing_params_repr(params, expected_repr):
    """Verify TextSizingParams.__repr__ output."""
    assert repr(params) == expected_repr


@pytest.mark.parametrize('params,text,expected_width', [
    # cases of static width=N values,
    (TextSizingParams(scale=2, width=1), 'climclam', 2),
    (TextSizingParams(scale=2, width=3), 'anything', 6),
    (TextSizingParams(scale=1, width=5), '', 5),
    (TextSizingParams(scale=3, width=1), 'x', 3),
    # and automatic width (width=0) values,
    (TextSizingParams(), '', 0),
    (TextSizingParams(), 'AB', 2),
    (TextSizingParams(), '中', 2),
    (TextSizingParams(scale=2), 'AB', 4),
    (TextSizingParams(scale=2), '中', 4),
    (TextSizingParams(scale=3), '', 0),
    (TextSizingParams(scale=7, width=7, numerator=15, denominator=15,
                      vertical_align=2, horizontal_align=2), 'x!yzzy', 49),
])
def test_text_sizing_width(params, text, expected_width):
    """Verify width using with both kinds of terminator."""
    # verify internal TextSizing.display_width() result,
    assert TextSizing(params, text, terminator='\x07').display_width() == expected_width
    assert TextSizing(params, text, terminator='\x1b\\').display_width() == expected_width
    seq1 = TextSizing(params, text, terminator='\x07').make_sequence()
    seq2 = TextSizing(params, text, terminator='\x1b\\').make_sequence()

    # verify round-trip
    ts_match1, ts_match2 = TEXT_SIZING_PATTERN.match(seq1), TEXT_SIZING_PATTERN.match(seq2)
    assert ts_match1 and ts_match2
    assert TextSizing.from_match(ts_match1) == TextSizing(params, text, terminator='\x07')
    assert TextSizing.from_match(ts_match2) == TextSizing(params, text, terminator='\x1b\\')

    # and external width(),
    assert width(seq1) == expected_width
    assert width(seq2) == expected_width

    # verify 'strict' does not raise ValueError
    width(seq1, control_codes='strict')
    width(seq2, control_codes='strict')

    # and verify 'ignore' measures only inner_text (does not parse scale or width)
    assert width(seq1, control_codes='ignore') == wcswidth(text)
    assert width(seq2, control_codes='ignore') == wcswidth(text)


@pytest.mark.parametrize('given_sequence,expected_text,expected_params,expected_width', [
    ('\x1b]66;s=2:w=2;AB\x07', 'AB', 's=2:w=2', 4),
    ('\x1b]66;s=2:w=2;\u4e2d\x07', '\u4e2d', 's=2:w=2', 4),
    ('\x1b]66;s=3:w=1;x\x07', 'x', 's=3:w=1', 3),
    ('\x1b]66;w=5;hello\x07', 'hello', 'w=5', 5),
    ('\x1b]66;s=2:w=3;anything\x07', 'anything', 's=2:w=3', 6),
    ('\x1b]66;w=3;x\x07', 'x', 'w=3', 3),
    ('\x1b]66;s=1;AB\x07', 'AB', '', 2),
    ('\x1b]66;s=2;AB\x07', 'AB', 's=2', 4),
    ('\x1b]66;s=2;中\x07', '中', 's=2', 4),
    ('\x1b]66;s=2;\x07', '', 's=2', 0),
    ('\x1b]66;s=1:w=1;\x07', '', 'w=1', 1),
    ('\x1b]66;w=2;A\x07', 'A', 'w=2', 2),
    ('\x1b]66;s=2:w=3;text\x1b\\', 'text', 's=2:w=3', 6),
])
def test_text_sizing_sequence(given_sequence, expected_text, expected_params, expected_width):
    """Verify parsing and measured width of raw OSC 66 sequence."""
    ts_match = TEXT_SIZING_PATTERN.match(given_sequence)
    assert ts_match is not None
    text_size = TextSizing.from_match(ts_match)
    assert text_size.params.make_sequence() == expected_params
    assert text_size.text == expected_text
    assert width(given_sequence, control_codes='parse') == expected_width
    assert width(given_sequence, control_codes='strict') == expected_width
    assert width(given_sequence, control_codes='ignore') == wcswidth(expected_text)


@pytest.mark.parametrize('text,expected', [
    ('\x1b]66;s=2:w=3:n=1:d=2:v=1:h=2;x!yzzy\x1b\\', 6),
    ('\x1b]66;s=2:w=3;anything\x07', 6),
    ('\x1b]66;w=3;x\x07', 3),
    ('\x1b]66;s=1:w=0;AB\x07', 2),
    ('\x1b]66;s=2:w=0;AB\x07', 4),
    ('\x1b]66;s=2:w=0;\u4e2d\x07', 4),  # '中'
    ('\x1b]66;s=1:w=0;\x07', 0),
    ('abc\x1b]66;w=3;x\x07def', 9),
    ('\x1b]66;w=2;A\x07\x1b]66;w=3;B\x07', 5),
    ('\x1b]66;s=2:w=3;text\x1b\\', 6),
    ('\x1b[31m\x1b]66;w=2;AB\x07\x1b[0m', 2),
    ('\x1b]66;;\x01\x07', 0),
])
def test_strings_with_text_sizing(text, expected):
    """Verify measured width strings containing OSC66."""
    assert width(text) == expected
    assert width(text, control_codes='strict') == expected


@pytest.mark.parametrize('text,expected', [
    ('\x1b]66;s=2;hello\x07', 'hello'),
    ('\x1b]66;s=2;hello\x1b\\', 'hello'),
    ('\x1b]66;;text\x07', 'text'),
    ('\x1b]66;s=3:w=2;\x07', ''),
    ('abc\x1b]66;w=2;XY\x07def', 'abcXYdef'),
    ('\x1b[31m\x1b]66;s=2;red\x07\x1b[0m', 'red'),
    ('\x1b]66;w=1;A\x07\x1b]66;w=1;B\x07', 'AB'),
])
def test_strip_strings_with_text_sizing(text, expected):
    assert strip_sequences(text) == expected


@pytest.mark.parametrize('text,expected_segs', [
    ('abc\x1b]66;s=2;hello\x07def', [('abc', False), ('\x1b]66;s=2;hello\x07', True), ('def', False)]),
    ('abc\x1b]66;s=2;n=1,d=2,w=3;hello\x1b\\def', [('abc', False), ('\x1b]66;s=2;n=1,d=2,w=3;hello\x1b\\', True), ('def', False)]),
])
def test_iter_sequences_text_sizing(text, expected_segs):
    assert list(iter_sequences(text)) == expected_segs


@pytest.mark.parametrize('text,start,end,expected', [
    ('\x1b]66;w=3;ABC\x07', 0, 3, '\x1b]66;w=3;ABC\x07'),
    ('\x1b]66;w=3;ABC\x07', 0, 2, '\x1b]66;w=2;AB\x07'),
    ('\x1b]66;w=3;ABC\x07', 1, 3, '\x1b]66;w=2;BC\x07'),
    ('ab\x1b]66;w=2;XY\x07cd', 0, 6, 'ab\x1b]66;w=2;XY\x07cd'),
    ('ab\x1b]66;w=2;XY\x07cd', 0, 3, 'ab\x1b]66;w=1;X\x07'),
    ('ab\x1b]66;w=2;XY\x07cd', 3, 6, '\x1b]66;w=1;Y\x07cd'),
    ('ab\x1b]66;w=2;XY\x07cd', 4, 6, 'cd'),
])
def test_clip_text_sizing_basic(text, start, end, expected):
    """Test basic support of clip() with text sizing sequence."""
    assert repr(clip(text, start, end)) == repr(expected)


@pytest.mark.parametrize('text,start,end,expected', [
    ('\x1b]66;s=2;ABC\x07', 0, 0, ''),
    ('\x1b]66;s=2;ABC\x07', 6, 6, ''),
    ('\x1b]66;s=2;ABC\x07', 0, 2, '\x1b]66;s=2;A\x07'),
    ('\x1b]66;s=2;ABC\x07', 0, 4, '\x1b]66;s=2;AB\x07'),
    ('\x1b]66;s=2;ABC\x07', 0, 6, '\x1b]66;s=2;ABC\x07'),
    ('\x1b]66;s=2;ABC\x07', 2, 6, '\x1b]66;s=2;BC\x07'),
    ('\x1b]66;s=2;ABC\x07', 4, 6, '\x1b]66;s=2;C\x07'),
])
def test_clip_text_sizing_scaled(text, start, end, expected):
    """Test support of clip() with scale=N arguments."""
    assert repr(clip(text, start, end)) == repr(expected)


@pytest.mark.parametrize('text,start,end,expected', [
    #  a   b   c
    # === === ===
    # 012 345 678
    # .
    # ..
    # *a*
    # *a* .
    # ... *b*
    # ... *b* .
    # ... *b* ..
    # ... *b* *c*
    ('\x1b]66;s=3;ABC\x07', 0, 0, ''),
    ('\x1b]66;s=3;ABC\x07', 0, 1, '.'),
    ('\x1b]66;s=3;ABC\x07', 0, 2, '..'),
    ('\x1b]66;s=3;ABC\x07', 0, 3, '\x1b]66;s=3;A\x07'),
    ('\x1b]66;s=3;ABC\x07', 0, 4, '\x1b]66;s=3;A\x07.'),
    ('\x1b]66;s=3;ABC\x07', 0, 5, '\x1b]66;s=3;A\x07..'),
    ('\x1b]66;s=3;ABC\x07', 0, 6, '\x1b]66;s=3;AB\x07'),
    ('\x1b]66;s=3;ABC\x07', 0, 7, '\x1b]66;s=3;AB\x07.'),
    ('\x1b]66;s=3;ABC\x07', 0, 8, '\x1b]66;s=3;AB\x07..'),
    ('\x1b]66;s=3;ABC\x07', 0, 9, '\x1b]66;s=3;ABC\x07'),
    ('\x1b]66;s=3;ABC\x07', 0, 10, '\x1b]66;s=3;ABC\x07'),
    #  a   b
    # === === ===
    # 012 345 678
    #  .             1, 2
    #  ..            1, 3
    #  .. .          1, 4
    #  .. ..         1, 5
    #  .. *b*        1, 6
    #  .. *b* .      1, 7
    #  .. *b* ..     1, 8
    #  .. *b* *c*    1, 9
    ('\x1b]66;s=3;ABC\x07', 1, 1, ''),
    ('\x1b]66;s=3;ABC\x07', 1, 2, '.'),
    ('\x1b]66;s=3;ABC\x07', 1, 3, '..'),
    ('\x1b]66;s=3;ABC\x07', 1, 4, '...'),
    ('\x1b]66;s=3;ABC\x07', 1, 5, '....'),
    ('\x1b]66;s=3;ABC\x07', 1, 6, '..\x1b]66;s=3;B\x07'),
    ('\x1b]66;s=3;ABC\x07', 1, 7, '..\x1b]66;s=3;B\x07.'),
    ('\x1b]66;s=3;ABC\x07', 1, 8, '..\x1b]66;s=3;B\x07..'),
    ('\x1b]66;s=3;ABC\x07', 1, 9, '..\x1b]66;s=3;BC\x07'),
    ('\x1b]66;s=3;ABC\x07', 1, 10, '..\x1b]66;s=3;BC\x07'),
    # two-thirds of string 'A' and half of string 'B' is fillchar
    # ('\x1b]66;s=3;ABC\x07', 2, 4, '..'),
    # half of string 'A' and all of string 'B'
    #  a   b
    # === === ===
    # 012 345 678
    #   .            2, 3
    #   . .          2, 4
    #   . ..         2, 5
    #   . *b*        2, 6
    #   . *b* .      2, 7
    #   . *b* ..     2, 8
    #   . *b* *c*    2, 9
    ('\x1b]66;s=3;ABC\x07', 2, 2, ''),
    ('\x1b]66;s=3;ABC\x07', 2, 3, '.'),
    ('\x1b]66;s=3;ABC\x07', 2, 4, '..'),
    ('\x1b]66;s=3;ABC\x07', 2, 5, '...'),
    ('\x1b]66;s=3;ABC\x07', 2, 6, '.\x1b]66;s=3;B\x07'),
    ('\x1b]66;s=3;ABC\x07', 2, 7, '.\x1b]66;s=3;B\x07.'),
    ('\x1b]66;s=3;ABC\x07', 2, 8, '.\x1b]66;s=3;B\x07..'),
    ('\x1b]66;s=3;ABC\x07', 2, 9, '.\x1b]66;s=3;BC\x07'),
    ('\x1b]66;s=3;ABC\x07', 2, 10, '.\x1b]66;s=3;BC\x07'),
    # and now 3:10, should be easy ...
    ('\x1b]66;s=3;ABC\x07', 3, 3, ''),
    ('\x1b]66;s=3;ABC\x07', 3, 4, '.'),
    ('\x1b]66;s=3;ABC\x07', 3, 5, '..'),
    ('\x1b]66;s=3;ABC\x07', 3, 6, '\x1b]66;s=3;B\x07'),
    ('\x1b]66;s=3;ABC\x07', 3, 7, '\x1b]66;s=3;B\x07.'),
    ('\x1b]66;s=3;ABC\x07', 3, 8, '\x1b]66;s=3;B\x07..'),
    ('\x1b]66;s=3;ABC\x07', 3, 9, '\x1b]66;s=3;BC\x07'),
    ('\x1b]66;s=3;ABC\x07', 3, 10, '\x1b]66;s=3;BC\x07'),
])
def test_clip_text_sizing_scaled_with_fillchar(text, start, end, expected):
    """Test support of clip() with scale=N and fillchar is needed to fill remainder."""
    assert repr(clip(text, start, end, fillchar='.')) == repr(expected)


def test_clip_simple_path_padding():
    """Simple-path clip with w=N larger than text length exercises padding loop."""
    # w=4 but only 1 grapheme 'X': 3 empty units are padded.
    # Clip window (0, 1) forces partial overlap, triggering
    # _text_sizing_clip_simple's padding branch.
    assert repr(clip('\x1b]66;w=4;X\x07', 0, 1)) == repr('\x1b]66;w=1;X\x07')


@pytest.mark.parametrize('text,start,end,expected', [
    # CR forces painter path; fully-visible text sizing sequence
    ('\r\x1b]66;w=2;XY\x07', 0, 3, '\x1b]66;w=2;XY\x07'),
    # CR painter path, text sizing partially clipped (first unit visible)
    ('\r\x1b]66;w=2;XY\x07', 0, 1, '\x1b]66;w=1;X\x07'),
    # BS forces painter path; text sizing fully visible
    ('ab\b\b\x1b]66;w=2;XY\x07', 0, 4, '\x1b]66;w=2;XY\x07'),
    # Painter path with partial text sizing overlap (exercises _text_sizing_clip_painter)
    ('\ra\x1b]66;s=2;BC\x07', 0, 3, 'a\x1b]66;s=2;B\x07'),
    # Painter path: text sizing scaled partial overlap with fillchar
    ('\r\x1b]66;s=3;ABC\x07', 1, 6, '  \x1b]66;s=3;B\x07'),
    # CSI movement + text sizing fully visible
    ('ab\x1b[2D\x1b]66;w=2;XY\x07', 0, 4, '\x1b]66;w=2;XY\x07'),
    # Painter path: text sizing entirely outside clip window (before start)
    ('\r\x1b]66;w=2;XY\x07', 2, 4, ''),
    # CR + text sizing with auto-width (w=0), partial overlap
    ('\ra\x1b]66;s=2;BC\x07', 0, 5, 'a\x1b]66;s=2;BC\x07'),
    # Painter path: padding when w=N has more units than graphemes
    ('\r\x1b]66;w=3;A\x07', 0, 2, '\x1b]66;w=2;A\x07'),
    # Painter path: text sizing with unit entirely before clip window (skip path)
    ('\r\x1b]66;s=2;ABCD\x07', 4, 8, '\x1b]66;s=2;CD\x07'),
])
def test_clip_text_sizing_painter(text, start, end, expected):
    """Test clip() with text sizing sequences in the cursor-movement (painter) path."""
    assert repr(clip(text, start, end)) == repr(expected)



"""Tests for ambiguous_width parameter."""

# 3rd party
import pytest

# local
import wcwidth

AMBIGUOUS_CHARS = [
    ('\u00a1', 'INVERTED_EXCLAMATION'),
    ('\u00a7', 'SECTION_SIGN'),
    ('\u00b0', 'DEGREE_SIGN'),
    ('\u00b1', 'PLUS_MINUS'),
    ('\u00d7', 'MULTIPLICATION'),
    ('\u00f7', 'DIVISION'),
    ('\u2460', 'CIRCLED_ONE'),
    ('\u2500', 'BOX_HORIZONTAL'),
    ('\u25a0', 'BLACK_SQUARE'),
    ('\u2605', 'BLACK_STAR'),
]


@pytest.mark.parametrize('char,name', AMBIGUOUS_CHARS)
def test_wcwidth_ambiguous_default(char, name):
    """Ambiguous characters return width 1 by default."""
    assert wcwidth.wcwidth(char) == 1


@pytest.mark.parametrize('char,name', AMBIGUOUS_CHARS)
def test_wcwidth_ambiguous_wide(char, name):
    """Ambiguous characters return width 2 when ambiguous_width=2."""
    assert wcwidth.wcwidth(char, ambiguous_width=2) == 2


def test_wcswidth_mixed_ambiguous_and_wide():
    """Mixed CJK and ambiguous characters."""
    text = '\u4e2d\u00b1'  # CJK (wide=2) + PLUS-MINUS (ambiguous)
    assert wcwidth.wcswidth(text) == 3
    assert wcwidth.wcswidth(text, ambiguous_width=2) == 4


def test_width_ambiguous():
    """Width() respects ambiguous_width parameter."""
    assert wcwidth.width('\u00b1') == 1
    assert wcwidth.width('\u00b1', ambiguous_width=2) == 2


def test_ljust_ambiguous():
    """Ljust respects ambiguous_width parameter."""
    text = '\u00b1'
    assert wcwidth.ljust(text, 4) == '\u00b1   '
    assert wcwidth.ljust(text, 4, ambiguous_width=2) == '\u00b1  '


def test_rjust_ambiguous():
    """Rjust respects ambiguous_width parameter."""
    text = '\u00b1'
    assert wcwidth.rjust(text, 4) == '   \u00b1'
    assert wcwidth.rjust(text, 4, ambiguous_width=2) == '  \u00b1'


def test_center_ambiguous():
    """Center respects ambiguous_width parameter."""
    text = '\u00b1'
    assert wcwidth.center(text, 5) == '  \u00b1  '
    assert wcwidth.center(text, 6, ambiguous_width=2) == '  \u00b1  '


def test_wrap_ambiguous():
    """Wrap respects ambiguous_width parameter."""
    text = '\u00b1' * 5  # 5 ambiguous characters
    assert wcwidth.wrap(text, 4) == ['\u00b1\u00b1\u00b1\u00b1', '\u00b1']
    assert wcwidth.wrap(text, 4, ambiguous_width=2) == ['\u00b1\u00b1', '\u00b1\u00b1', '\u00b1']


def test_wide_not_affected_by_ambiguous():
    """Wide characters remain wide regardless of ambiguous_width."""
    cjk = '\u4e2d'  # CJK character (always wide)
    assert wcwidth.wcwidth(cjk) == 2
    assert wcwidth.wcwidth(cjk, ambiguous_width=2) == 2
    assert wcwidth.wcwidth(cjk, ambiguous_width=1) == 2
